# !/usr/bin/python
# coding=utf-8
try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)


class Macro:
    """Assign macro functions to hotkeys.

    Example:
    class MSlots(Macro):
            '''A class that inherits from `Macro` and holds any macro functions.
            '''
            @staticmethod
            def m_back_face_culling():
                    '''Toggle Back-Face Culling.
                    '''
                    sel = pm.ls(selection=True)
                    if sel:
                            currentPanel = getPanel(withFocus=True)
                            state = pm.polyOptions(sel, query=True, wireBackCulling=True)[0]

                            if not state:
                                    pm.polyOptions(sel, gl=True, wireBackCulling=True)
                                    Macros.setWireframeOnShadedOption(currentPanel, 0)
                                    pm.inViewMessage(status_message="Back-Face Culling is now <hl>OFF</hl>.>", pos='topCenter', fade=True)
                            else:
                                    pm.polyOptions(sel, gl=True, backCulling=True)
                                    Macros.setWireframeOnShadedOption(currentPanel, 1)
                                    pm.inViewMessage(status_message="Back-Face Culling is now <hl>ON</hl>.", pos='topCenter', fade=True)
                    else:
                            print(" Warning: Nothing selected. ")

    #call the `set_macros` function to set a macro for functions you defined in `MSlots`.
    MSlots.set_macros(macros={
            'm_back_face_culling': {'k':'1', 'cat':'Display'},
    }
    """

    @classmethod
    def set_macros(cls, *args):
        """Extends `set_macro` to accept a list of strings representing positional and keyword arguments.

        Parameters:
                *args (str): A variable number of strings, each containing the arguments for a single macro. Each string
                                should be in the format "<macro name>, <positional arg1>, <positional arg2>, ..., <keyword arg1>=<value1>,
                                <keyword arg2>=<value2>, ..."
        Example:
                set_macros('m_back_face_culling, key=1, cat=Display', 'm_smooth_preview, key=2, cat=Display') #Calls `set_macro` with the parsed arguments for each macro in `args`.
        """
        for string in args:
            cls.call_with_input(cls.set_macro, string)

    @staticmethod
    def call_with_input(func, input_string):
        """Parses an input string into positional and keyword arguments, and
        calls the given function with those arguments.

        Parameters:
                func (callable): The function to call.
                input_string (str): The input string containing the arguments.

        Returns:
                The result of calling `func` with the parsed arguments.
        """
        args, kwargs = [], {}
        for i in input_string.split(","):
            try:
                key, value = i.split("=")
                kwargs[key.strip()] = value.strip()
            except ValueError:
                args.append(i.strip())

        return func(*args, **kwargs)

    @classmethod
    def set_macro(
        cls, name, key=None, cat=None, ann=None, default=False, delete_existing=True
    ):
        """Sets a default runtime command with a keyboard shortcut.

        Parameters:
                name (str): The command name you provide must be unique. (alphanumeric characters, or underscores)
                cat (str): catagory - Category for the command.
                ann (str): annotation - Description of the command.
                key (str): keyShortcut - Specify what key is being set.
                                        key modifier values are set by adding a '+' between chars. ie. 'sht+z'.
                                        modifiers:
                                                alt, ctl, sht
                                        additional valid keywords are:
                                                Up, Down, Right, Left,
                                                Home, End, Page_Up, Page_Down, Insert
                                                Return, Space
                                                F1 to F12
                                                Tab (Will only work when modifiers are specified)
                                                Delete, Backspace (Will only work when modifiers are specified)
                default (bool): Indicate that this run time command is a default command. Default run time commands will not be saved to preferences.
                delete_existing = Delete any existing (non-default) runtime commands of the given name.
        """
        command = f"if 'm_slots' not in globals(): from {cls.__module__} import {cls.__name__}; global m_slots; m_slots = {cls.__name__}();\nm_slots.{name}();"

        if not ann:  # if no ann is given, try using the method's docstring.
            method = getattr(cls, name)
            ann = method.__doc__.split("\n")[0]  # use only the first line.

        if pm.runTimeCommand(name, exists=True):
            if pm.runTimeCommand(name, query=True, default=True):
                return  # can not delete default runtime commands.
            elif (
                delete_existing
            ):  # delete any existing (non-default) runtime commands of that name.
                pm.runTimeCommand(name, edit=True, delete=True)

        try:  # set runTimeCommand
            pm.runTimeCommand(
                name,
                annotation=ann,
                category=cat,
                command=command,
                default=default,
            )
        except RuntimeError as error:
            print("# Error: {}: {} #".format(__file__, error))
            return error

        # set command
        nameCommand = pm.nameCommand(
            "{0}Command".format(name),
            annotation=ann,
            command=name,
        )

        # set hotkey
        # modifiers
        ctl = False
        alt = False
        sht = False
        for char in key.split("+"):
            if char == "ctl":
                ctl = True
            elif char == "alt":
                alt = True
            elif char == "sht":
                sht = True
            else:
                key = char

        # print(name, char, ctl, alt, sht)
        pm.hotkey(
            keyShortcut=key, name=nameCommand, ctl=ctl, alt=alt, sht=sht
        )  # set only the key press.


# --------------------------------------------------------------------------------------------

if __name__ == "__main__":
    pass

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------

"""
#create wrapper
mel.createMelWrapper(method)

#set command
pm.nameCommand('name', annotation='', command=<>)
pm.hotkey(key='1', altModifier=True, name='name')


#clear keyboard shortcut
pm.hotkey(keyShortcut=key, name='', releaseName='', ctl=ctl, alt=alt, sht=sht) #unset the key press name and releaseName.


#query runTimeCommand
if pm.runTimeCommand('name', exists=True):


#delete runTimeCommand
pm.runTimeCommand('name', edit=True, delete=True)


#set runTimeCommand
pm.runTimeCommand(
            'name',
            annotation=string,
            category=string,
            categoryArray,
            command=script,
            commandArray,
            commandLanguage=string,
            default=boolean,
            defaultCommandArray,
            delete,
            exists,
            hotkeyCtx=string,
            image=string,
            keywords=string,
            annotation=string,
            longAnnotation=string,
            numberOfCommands,
            numberOfDefaultCommands,
            numberOfUserCommands,
            plugin=string,
            save,
            showInHotkeyEditor=boolean,
            tags=string,
            userCommandArray,
)

-annotation(-ann) string createqueryedit 
        Description of the command.

-category(-cat) string createqueryedit  
        Category for the command.

-categoryArray(-caa) query          
        Return all the run time command categories.

-command(-c) script createqueryedit     
        Command to be executed when runTimeCommand is invoked.

-commandArray(-ca) query                
        Returns an string array containing the names of all the run time commands.

-commandLanguage(-cl) string createqueryedit
        In edit or create mode, this flag allows the caller to choose a scripting language for a command passed to the "-command" flag. If this flag is not specified, then the callback will be assumed to be in the language from which the runTimeCommand command was called. In query mode, the language for this runTimeCommand is returned. The possible values are "mel" or "python".

-default(-d) boolean createquery        
        Indicate that this run time command is a default command. Default run time commands will not be saved to preferences.

-defaultCommandArray(-dca) query                
        Returns an string array containing the names of all the default run time commands.

-delete(-del) edit              
        Delete the specified user run time command.

-exists(-ex) create                 
        Returns true|false depending upon whether the specified object exists. Other flags are ignored.

-hotkeyCtx(-hc) string createqueryedit  
        hotkey Context for the command.

-image(-i) string createqueryedit   
        Image filename for the command.

-keywords(-k) string createqueryedit        
        Keywords for the command. Used for searching for commands in Type To Find. When multiple keywords, use ; as a separator. (Example: "keyword1;keyword2")

-annotation(-annotation) string createqueryedit     
        Label for the command.

-longAnnotation(-la) string createqueryedit 
        Extensive, multi-line description of the command. This will show up in Type To Finds more info page in addition to the annotation.

-numberOfCommands(-nc) query            
        Return the number of run time commands.

-numberOfDefaultCommands(-ndc) query            
        Return the number of default run time commands.

-numberOfUserCommands(-nuc) query           
        Return the number of user run time commands.

-plugin(-p) string createqueryedit          
        Name of the plugin this command requires to be loaded. This flag wraps the script provided into a safety check and automatically loads the plugin referenced on execution if it hasn't been loaded. If the plugin fails to load, the command won't be executed.

-save(-s) edit                          
        Save all the user run time commands.

-showInHotkeyEditor(-she) boolean createqueryedit       
        Indicate that this run time command should be shown in the Hotkey Editor. Default value is true.

-tags(-t) string createqueryedit    
        Tags for the command. Used for grouping commands in Type To Find. When more than one tag, use ; as a separator. (Example: "tag1;tag2")

-userCommandArray(-uca) query           
        Returns an string array containing the names of all the user run time commands.
"""


# --------------------------------------------------------------------------------------------
# deprecated:
# --------------------------------------------------------------------------------------------


## command = self.formatSource(name, removeTabs=1) #remove 1 tab space.
# def formatSource(self, cmd, removeTabs=0):
#       '''Return the text of the source code for an object.
#       The source code is returned as a single string.
#       Removes lines containing '@' or 'def ' ie. @staticmethod.

#       Parameters:
#           cmd = module, class, method, function, traceback, frame, or code object.
#           removeTabs (int): remove x instances of '\t' from each line.

#       Returns:
#           A Multi-line string.
#       '''
#       from inspect import getsource
#       source = getsource(getattr(Macros, cmd))

#       l = [s.replace('\t', '', removeTabs) for s in source.split('\n') if s and not '@' in s]
#       call = [s.replace('\t', '', removeTabs).lstrip('def ').rstrip(':') for s in source.split('\n') if 'def ' in s]
#       return '\n'.join(l)+'\n\n'+call[0]


# string $image_plane[] = `ls -exactType imagePlane`;
# for ($object in $image_plane){
#   if (`getAttr ($object+".displayMode")` != 2){
#       setAttr ($object+".displayMode") 2;
#       grid -toggle 1;
#       inViewMessage -status_message "Grid is now <hl>ON</hl>.\\n<hl>F1</hl>"  -fade -position topCenter;
#   }else{
#       setAttr ($object+".displayMode") 0;
#       grid -toggle 0;
#       inViewMessage -status_message "Grid is now <hl>OFF</hl>.\\n<hl>F1</hl>"  -fade -position topCenter;
#   }
# }


#       global int $toggleFrame_;

#       string $selection[] = `ls -selection`;

#       $mode = `selectMode -query -component`;
#       $maskVertex = `selectType -query -vertex`;
#       $maskEdge = `selectType -query -edge`;
#       $maskFacet = `selectType -query -facet`;

#       if (size($selection)==0)
#           {
#           viewFit -allObjects;
#           }

#       if ($mode==1 && $maskVertex==1 && size($selection)!=0)
#           {
#           if (size($selection)>1)
#               {
#               if ($toggleFrame_ == !1)
#                   {
#                   viewFit -fitFactor .65;
#                   $toggleFrame_ = 1;
#                   print ("frame vertices "+$toggleFrame_+"\\n");
#                   }
#               else
#                   {
#                   viewFit -fitFactor .10;
#                   //viewSet -previousView;
#                   $toggleFrame_ = 0;
#                   print ("frame vertices "+$toggleFrame_+"\\n");
#                   }
#               }
#           else
#               {
#               if ($toggleFrame_ == !1)
#                   {
#                   viewFit -fitFactor .15;
#                   $toggleFrame_ = 1;
#                   print ("frame vertex "+$toggleFrame_+"\\n");
#                   }
#               else
#                   {
#                   viewFit -fitFactor .01;
#                   //viewSet -previousView;
#                   $toggleFrame_ = 0;
#                   print ("frame vertex "+$toggleFrame_+"\\n");
#                   }
#               }
#           }
#       if ($mode==1 && $maskEdge==1 && size($selection)!=0)
#           {
#           if ($toggleFrame_ == !1)
#               {
#               viewFit -fitFactor .3;
#               $toggleFrame_ = 1;
#               print ("frame edge "+$toggleFrame_+"\\n");
#               }
#           else
#               {
#               viewFit -fitFactor .9;
#               //viewSet -previousView;
#               $toggleFrame_ = 0;
#               print ("frame edge "+$toggleFrame_+"\\n");
#               }
#           }
#       if ($mode==1 && $maskFacet==1)
#           {
#           if ($toggleFrame_ == !1)
#               {
#               viewFit -fitFactor .9;
#               $toggleFrame_ = 1;
#               print ("frame facet "+$toggleFrame_+"\\n");
#               }
#           else
#               {
#               viewFit -fitFactor .45;
#               //viewSet -previousView;
#               $toggleFrame_ = 0;
#               print ("frame facet "+$toggleFrame_+"\\n");
#               }
#           }
#       else if ($mode==0  && size($selection)!=0)
#           {
#           if ($toggleFrame_ == !1)
#               {
#               viewFit -fitFactor .99;
#               $toggleFrame_ = 1;
#               print ("frame object "+$toggleFrame_+"\\n");
#               }
#           else
#               {
#               viewFit -fitFactor .65;
#               //viewSet -previousView;
#               $toggleFrame_ = 0;
#               print ("frame object "+$toggleFrame_+"\\n");
#               }
#           }


# string $currentPanel = `getPanel -withFocus`;
# $mode = `displayPref -query -wireframeOnShadedActive`;

# if ($mode=="none")
#   {
#   displayPref -wireframeOnShadedActive "reduced";
#   setWireframeOnShadedOption 1 $currentPanel;
#   inViewMessage -status_message "<hl>Wireframe-on-selection</hl> is now <hl>Full</hl>.\\n<hl>3</hl>"  -fade -position topCenter;
#   }
# if ($mode=="reduced")
#   {
#   displayPref -wireframeOnShadedActive "full";
#   setWireframeOnShadedOption 0 $currentPanel;
#   inViewMessage -status_message "<hl>Wireframe-on-selection</hl> is now <hl>Reduced</hl>.\\n<hl>3</hl>"  -fade -position topCenter;
#   }
# if ($mode=="full")
#   {
#   displayPref -wireframeOnShadedActive "none";
#   setWireframeOnShadedOption 0 $currentPanel;
#   inViewMessage -status_message "<hl>Wireframe-on-selection</hl> is now <hl>OFF</hl>.\\n<hl>3</hl>" -fade -position topCenter;
#   }


# //xray all except selected
# string $scene[] = `ls -visible -flatten -dag -noIntermediate -type surfaceShape`;
# string $selection[] = `ls -selection -dagObjects -shapes`;
# for ($object in $scene)
#   {
#   if (!stringArrayContains ($object, $selection))
#       {
#       int $state[] = `displaySurface -query -xRay $object`;
#       displaySurface -xRay ( !$state[0] ) $object;
#       }
#   }


# mel.eval('''
# string $currentPanel = `getPanel -withFocus`;
# string $state = `modelEditor -query -displayAppearance $currentPanel`;
# string $displayTextures = `modelEditor -query -displayTextures $currentPanel`;
# if(`modelEditor -exists $currentPanel`)
#   {
#   if($state != "wireframe" && $displayTextures == false)
#     {
#       modelEditor -edit -displayAppearance smoothShaded -activeOnly false -displayTextures true $currentPanel;
#       inViewMessage -status_message "modelEditor -smoothShaded <hl>true</hl> -displayTextures <hl>true</hl>.\\n<hl>5</hl>"  -fade -position topCenter;
#       }
#   if($state == "wireframe" && $displayTextures == true)
#     {
#       modelEditor -edit -displayAppearance smoothShaded -activeOnly false -displayTextures false $currentPanel;
#       inViewMessage -status_message "modelEditor -smoothShaded <hl>true</hl> -displayTextures <hl>false</hl>.\\n<hl>5</hl>"  -fade -position topCenter;
#       }
#   if($state != "wireframe" && $displayTextures == true)
#     {
#       modelEditor -edit -displayAppearance wireframe -activeOnly false $currentPanel;
#       inViewMessage -status_message "modelEditor -wireframe <hl>true</hl>.\\n<hl>5</hl>"  -fade -position topCenter;
#       }
#   }
# ''')


# SelectMultiComponentMask;
# inViewMessage -status_message "<hl>Multi-Component Selection Mode</hl>\\n Mask is now <hl>ON</hl>.\\n<hl>F4</hl>"  -fade -position topCenter;


# //paste and then re-name object removing keyword 'pasted'
# cutCopyPaste "paste";
# {
# string $pasted[] = `ls "pasted__*"`;
# string $object;
# for ( $object in $pasted )
# {
# string $elements[];
# // The values returned by ls may be full or partial dag
# // paths - when renaming we only want the actual
# // object name so strip off the leading dag path.
# //
# tokenize( $object, "|", $elements );
# string $stripped = $elements[ `size $elements` - 1 ];
# // Remove the 'pasted__' suffix from the name
# //
# $stripped = `substitute "pasted__" $stripped ""`;
# // When renaming a transform its shape will automatically be
# // be renamed as well. Use catchQuiet here to ignore errors
# // when trying to rename the child shape a second time.
# //
# catchQuiet(`evalEcho("rename " + $object + " " + $stripped)`);
# }
# };
# //alternative: edit the cutCopyPaste.mel
# //REMOVE the line "-renameAll" so the sub-nodes won't get renamed at all
# //REMOVE the -renamingPrefix "paste_" line
# //and instead write the line -defaultNamespace


# $mode = `selectMode -query -component`;
# if ($mode==0)
#   {
#   changeSelectMode -component;
#   }

# $maskVertex = `selectType -query -vertex`;
# $maskEdge = `selectType -query -edge`;
# $maskFacet = `selectType -query -facet`;

# if ($maskEdge==0 && $maskFacet==1)
#   {
#   selectType -vertex true;
#   inViewMessage -status_message "<hl>Vertex</hl> Mask is now <hl>ON</hl>.\\n<hl>F4</hl>"  -fade -position topCenter;
#   }
# if ($maskVertex==1 && $maskFacet==0)
#   {
#   selectType -edge true;
#   inViewMessage -status_message "<hl>Edge</hl> Mask is now <hl>ON</hl>.\\n<hl>F4</hl>"  -fade -position topCenter;
#   }
# if ($maskVertex==0 && $maskEdge==1)
#   {
#   selectType -facet true;
#   inViewMessage -status_message "<hl>Facet</hl> Mask is now <hl>ON</hl>.\\n<hl>F4</hl>"  -fade -position topCenter;
#   }


# // added this expression to fix 'toggleMainMenubar function not found' error
# if (`menu -q -ni MayaWindow|HotBoxControlsMenu` == 0) {setParent -m MayaWindow|HotBoxControlsMenu;source HotboxControlsMenu;};

# //toggle panel menus
# string $panels[] = `getPanel -allPanels`;
# int $state = `panel -query -menuBarVisible $panels[0]`;
# for ($panel in $panels)
# {
#   // int $state = `panel -query -menuBarVisible $panel`;
#   panel -edit -menuBarVisible (!$state) $panel;
# }
# //toggle main menubar
# toggleMainMenubar (!$state);

# //toggle titlebar
# window -edit -titleBar (!$state) $gMainWindow;

# // //toggle fullscreen mode //working but issues with windows resizing on toggle
# // int $inFullScreenMode = `optionVar -q "workspacesInFullScreenUIMode"`;
# // int $inZoomInMode = `optionVar -q "workspacesInZoomInUIMode"`;
# // // enter full screen mode only if the zoom-in mode is not active.
# // if(!$inZoomInMode)
# // {
# //    string $panelWithFocus = `getPanel -withFocus`;
# //    string $parentControl = `workspaceLayoutManager -parentWorkspaceControl $panelWithFocus`;
# //    int $isFloatingPanel = `workspaceControl -q -floating $parentControl`;

# //    if(!$isFloatingPanel)
# //    {
# //        if($inFullScreenMode)
# //        {
# //        //come out of fullscreen mode
# //        workspaceLayoutManager -restoreMainWindowControls;
# //        }
# //        else
# //        {
# //            // enter fullscreen mode
# //            workspaceLayoutManager -collapseMainWindowControls $parentControl true;
# //        }
# //    optionVar -iv "workspacesInFullScreenUIMode" (!$inFullScreenMode);
# //    }
# // }



# !/usr/bin/python
# coding=utf-8
import os
import unittest
import inspect
import pymel.core as pm
from mayatk import XformUtils


# sfr = pm.melGlobals['cmdScrollFieldReporter']
# pm.cmdScrollFieldReporter(sfr, edit=1, clear=1)


class Main(unittest.TestCase):
    """Main test class."""

    def perform_test(self, cases):
        """Execute the test cases."""
        for case in cases:
            if isinstance(case, str):
                expression = case
                expected_result = cases[case]
                method_name = str(expression).split("(")[0]
            else:
                result, expected_result = case
                method_name = result.__class__.__name__
                expression = None

            try:
                path = os.path.abspath(inspect.getfile(eval(method_name)))
            except (TypeError, IOError):
                path = ""

            if expression:
                result = eval(expression)

            self.assertEqual(
                result,
                expected_result,
                f"\n\n# Error: {path}\n#\tCall: {method_name})\n#\tExpected {type(expected_result)}: {expected_result}\n#\tReturned {type(result)}: {result}",
            )


class XformUtils_test(Main, XformUtils):
    """ """

    # Tear down the any previous test by creating a new scene:
    pm.mel.file(new=True, force=True)

    # assemble the test scene:
    if not pm.objExists("cube1"):
        cube1 = pm.polyCube(
            width=5,
            height=5,
            depth=5,
            subdivisionsX=1,
            subdivisionsY=1,
            subdivisionsZ=1,
            name="cube1",
        )

    if not pm.objExists("cube2"):
        cube2 = pm.polyCube(
            width=2,
            height=4,
            depth=8,
            subdivisionsX=3,
            subdivisionsY=3,
            subdivisionsZ=3,
            name="cube2",
        )

    if not pm.objExists("sph"):
        sph = pm.polySphere(radius=5, subdivisionsX=12, subdivisionsY=12, name="sph")

    def test_moveTo(self):
        """ """
        self.assertEqual(self.move_to("cube1", "cube2"), None)

    def test_dropToGrid(self):
        """ """
        self.assertEqual(
            self.drop_to_grid(
                "cube1",
                align="Min",
                origin=True,
                center_pivot=True,
                freeze_transforms=True,
            ),
            None,
        )

    def test_resetTranslation(self):
        """ """
        self.assertEqual(self.reset_translation("cube1"), None)

    def test_setTranslationToPivot(self):
        """ """
        self.assertEqual(self.set_translation_to_pivot("cube1"), None)

    def test_alignPivotToSelection(self):
        """ """
        self.assertEqual(self.align_pivot_to_selection("cube1", "cube2"), None)

    def test_aimObjectAtPoint(self):
        """ """
        self.assertEqual(
            self.aim_object_at_point(["cube1", "cube2"], (0, 15, 15)), None
        )

    def test_rotateAxis(self):
        """ """
        self.assertEqual(self.rotate_axis(["cube1", "cube2"], (0, 15, 15)), None)

    def test_getOrientation(self):
        """ """
        self.assertEqual(
            self.get_orientation("cube1"), ([1, 0, 0], [0, 1, 0], [0, 0, 1])
        )

    def test_getDistanceBetweenTwoObjects(self):
        """ """
        self.drop_to_grid(["cube1", "cube2"], origin=True, center_pivot=True)
        pm.move("cube2", 0, 0, 15)

        self.assertEqual(self.get_dist_between_two_objects("cube1", "cube2"), 15)

    def test_getCenterPoint(self):
        """ """
        self.assertEqual(self.get_center_point("sph"), (0, 0, 0))
        self.assertEqual(self.get_center_point("sph.vtx[*]"), (0, 0, 0))

    def test_getBoundingBox(self):
        """ """
        self.assertEqual(self.get_bounding_box("sph", "size"), (10, 10, 10))

    def test_sortByBoundingBoxValue(self):
        """ """
        self.assertEqual(
            str(self.sort_by_bounding_box_value(["sph.vtx[0]", "sph.f[0]"])),
            "[MeshFace('sphShape.f[0]'), MeshVertex('sphShape.vtx[0]')]",
        )

    def test_matchScale(self):
        """ """
        self.assertEqual(
            self.match_scale("cube1", "cube2", scale=False),
            [1.3063946090989371, 0.539387725343009, 0.539387708993454],
        )

    def test_snap3PointsTo3Points(self):
        """ """
        # self.assertEqual(self.align_using_three_points(), None)

    def test_isOverlapping(self):
        """ """
        # self.assertEqual(self.is_overlapping(), None)

    def test_alignVertices(self):
        """ """
        # self.assertEqual(self.align_vertices(), None)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main(exit=False)


# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

# """
# def test_(self):
#   '''
#   '''
#   self.perform_test({
#       # "self.": '',
#   })


# def test_(self):
#   '''
#   '''
#   self.perform_test({
#       # "self.": '',
#   })


# def test_(self):
#   '''
#   '''
#   self.perform_test({
#       # "self.": '',
#   })


# def test_(self):
#   '''
#   '''
#   self.perform_test({
#       # "self.": '',
#   })
# """

# # Deprecated ---------------------



# !/usr/bin/python
# coding=utf-8
try:
    import pymel.core as pm
except ModuleNotFoundError as error:
    print(__file__, error)
try:
    import maya.OpenMaya as om
except ModuleNotFoundError as error:
    print(__file__, error)

# from this package:
from mayatk.core_utils import CoreUtils
from mayatk.xform_utils import XformUtils
from mayatk.node_utils import NodeUtils


class ExplodedView:
    exploded_objects = []

    @staticmethod
    def calculate_repulsive_force(centroid1, size1, centroid2, size2, scale=0.05):
        """Calculates the repulsive force between two objects based on their centroids and sizes.

        Parameters:
            centroid1 (tuple): A tuple of floats representing the centroid of the first object in 3D space.
            size1 (float): The size of the first object.
            centroid2 (tuple): A tuple of floats representing the centroid of the second object in 3D space.
            size2 (float): The size of the second object.
            scale (float, optional): The scaling factor to apply to the force magnitude. Defaults to 0.05.

        Returns:
            om.MVector: The force vector resulting from the repulsion.
        """
        distance = (
            (centroid1[0] - centroid2[0]) ** 2
            + (centroid1[1] - centroid2[1]) ** 2
            + (centroid1[2] - centroid2[2]) ** 2
        ) ** 0.5

        epsilon = 1e-6
        force_magnitude = (size1 * size2) / ((distance * distance) + epsilon)
        force_magnitude *= scale

        force_vector = om.MVector(
            centroid1[0] - centroid2[0],
            centroid1[1] - centroid2[1],
            centroid1[2] - centroid2[2],
        )
        force_vector.normalize()
        force_vector *= force_magnitude

        return force_vector

    @classmethod
    def arrange_objects(
        cls, nodes, convergence_threshold=1e-4, max_iterations=1000, max_movement=1.0
    ):
        """Arranges a list of objects in 3D space to avoid overlap.

        Parameters:
            nodes (list): A list of objects to arrange.
            convergence_threshold (float, optional): The threshold at which to consider the system as having converged.
            Defaults to 1e-4.
            max_iterations (int, optional): The maximum number of iterations to run before stopping.
            Defaults to 10000.
            max_movement (float, optional): The maximum distance an object can move during a single iteration.
            Defaults to 1.0.

        Returns:
            int: The number of iterations required for the system to converge.
        """
        iteration_count = 0
        converged = False

        node_data = [
            XformUtils.get_bounding_box(node, "center|maxsize") for node in nodes
        ]

        while not converged and iteration_count < max_iterations:
            total_system_force = om.MVector(0, 0, 0)

            for idx1, node1 in enumerate(nodes):
                total_force = om.MVector(0, 0, 0)

                for idx2, node2 in enumerate(nodes):
                    if node1 != node2:
                        repulsive_force = cls.calculate_repulsive_force(
                            node_data[idx1][0],
                            node_data[idx1][1],
                            node_data[idx2][0],
                            node_data[idx2][1],
                        )
                        total_force += repulsive_force

                movement_vector = om.MVector(
                    total_force.x, total_force.y, total_force.z
                )

                if movement_vector.length() > max_movement:
                    movement_vector.normalize()
                    movement_vector *= max_movement

                pm.move(
                    movement_vector.x,
                    movement_vector.y,
                    movement_vector.z,
                    node1,
                    relative=True,
                )
                total_system_force += movement_vector

                # Update centroid
                new_centroid = [
                    node_data[idx1][0][i] + movement_vector[i] for i in range(3)
                ]
                node_data[idx1] = (new_centroid, node_data[idx1][1])

            if total_system_force.length() < convergence_threshold:
                converged = True

            iteration_count += 1

        cls.exploded_objects.append(nodes)
        return iteration_count

    def explode_selected(self):
        """Explode selected"""
        selection = NodeUtils.get_unique_children(pm.ls(sl=True))
        for obj in selection:
            if obj.hasAttr("original_position"):
                selection.remove(obj)
                continue
            pos = pm.xform(obj, query=True, translation=True, worldSpace=True)
            NodeUtils.set_node_attributes(obj, original_position=pos)

        iterations = self.arrange_objects(selection)

    def un_explode_selected(self):
        """Un-explode selected"""
        selection = NodeUtils.get_unique_children(pm.ls(sl=True))
        for obj in selection:
            if pm.attributeQuery("original_position", node=obj, exists=True):
                pos = pm.getAttr(obj.original_position)
                pm.move(pos[0], pos[1], pos[2], obj, absolute=True)
                pm.deleteAttr(obj, attribute="original_position")

    def un_explode_all(self):
        """Un-explode all"""
        all_objects_with_original_position = pm.ls("*.original_position")
        for obj_attr in all_objects_with_original_position:
            obj = obj_attr.node()
            pos = pm.getAttr(obj.original_position)
            pm.move(pos[0], pos[1], pos[2], obj, absolute=True)
            pm.deleteAttr(obj, attribute="original_position")

    def toggle_explode(self):
        """Toggle explode"""
        selection = NodeUtils.get_unique_children(pm.ls(sl=True))
        if selection:
            if pm.attributeQuery("original_position", node=selection[0], exists=True):
                self.un_explode_selected()
            else:
                self.explode_selected()


class ExplodedViewSlots(ExplodedView):
    def b000(self):
        """Explode button"""
        self.explode_selected()

    def b001(self):
        """Un-explode selected button"""
        self.un_explode_selected()

    def b002(self):
        """Un-explode all button"""
        self.un_explode_all()

    def b003(self):
        """Toggle Exlode"""
        self.toggle_explode()


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import os
    from uitk import Switchboard

    parent = CoreUtils.get_main_window()
    ui_file = os.path.join(os.path.dirname(__file__), "exploded_view.ui")
    sb = Switchboard(parent, ui_location=ui_file, slot_location=ExplodedViewSlots)

    sb.current_ui.set_attributes(WA_TranslucentBackground=True)
    sb.current_ui.set_flags(
        Tool=True, FramelessWindowHint=True, WindowStaysOnTopHint=True
    )
    sb.current_ui.set_style(theme="dark", style_class="translucentBgWithBorder")

    sb.current_ui.show(pos="screen", app_exec=True)

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)
import pythontk as ptk

# from this package:
from mayatk import node_utils


class MatUtils(object):
    """ """

    @staticmethod
    def get_mats(objs):
        """Returns the set of materials assigned to a given list of objects or components.

        Parameters:
            objs (list): The objects or components to retrieve the material from.

        Returns:
            set: The set of materials assigned to the objects or components.
        """
        # Initialize an empty set to store the materials
        mats = set()

        for obj in pm.ls(objs, flatten=True):
            shading_grps = []

            if isinstance(obj, pm.MeshFace):
                all_shading_grps = pm.ls(type="shadingEngine")
                shading_grps = [
                    sg for sg in all_shading_grps if pm.sets(sg, isMember=obj)
                ]

                if not shading_grps:
                    pm.hyperShade(obj, shaderNetworksSelectMaterialNodes=True)
                    mats_ = pm.ls(pm.selected(), materials=True)
                    mats.update(mats_)
            else:
                # Check if the object has a shape node
                shape = None
                if hasattr(obj, "getShape"):
                    shape = obj.getShape()

                if shape:
                    shading_grps = pm.listConnections(shape, type="shadingEngine")

            for shading_grp in shading_grps:
                materials = pm.ls(
                    pm.listConnections(f"{shading_grp}.surfaceShader"), materials=True
                )
                mats.update(materials)

        return mats

    @staticmethod
    def get_scene_mats(inc=[], exc=[]):
        """Retrieves all materials from the current scene, optionally including or excluding certain objects.

        Parameters:
            inc (str/int/obj/list, optional): The objects to include in the search. Supports using the '*' operator for pattern matching. Defaults to [].
            exc (str/int/obj/list, optional): The objects to exclude from the search. Supports using the '*' operator for pattern matching. Defaults to [].

        Returns:
            list: A list of materials in the scene.
        """
        matList = pm.ls(mat=1, flatten=1)

        # convert to dictionary to filter material names and types.
        d = {m.name(): pm.nodeType(m) for m in matList}
        filtered = ptk.filter_dict(d, inc, exc, keys=True, values=True)

        # use the filtered results to reconstruct a filtered list of actual materials.
        return [m for m in matList if m.name() in filtered]

    @staticmethod
    def get_fav_mats():
        """Retrieves the list of favorite materials in Maya.

        Returns:
            list: A list of favorite materials.
        """
        import os.path
        import maya.app.general.tlfavorites as _fav

        version = pm.about(version=True).split(" ")[-1]  # Get the Maya version year
        path = os.path.expandvars(
            f"%USERPROFILE%/Documents/maya/{version}/prefs/renderNodeTypeFavorites"
        )
        renderNodeTypeFavorites = _fav.readFavorites(path)
        materials = [i for i in renderNodeTypeFavorites if "/" not in i]
        del _fav

        return materials

    @staticmethod
    def create_mat(mat_type, prefix="", name=""):
        """Creates a material based on the provided type or a random Lambert material if 'mat_type' is 'random'.

        Parameters:
            mat_type (str): The type of the material, e.g. 'lambert', 'blinn', or 'random' for a random Lambert material.
            prefix (str, optional): An optional prefix to append to the material name. Defaults to "".
            name (str, optional): The name of the material. Defaults to "".

        Returns:
            obj: The created material.
        """
        import random

        if mat_type == "random":
            rgb = [
                random.randint(0, 255) for _ in range(3)
            ]  # Generate a list containing 3 values between 0-255
            name = "{}{}_{}_{}_{}".format(
                prefix, name, str(rgb[0]), str(rgb[1]), str(rgb[2])
            )
            mat = pm.shadingNode("lambert", asShader=True, name=name)
            convertedRGB = [round(float(v) / 255, 3) for v in rgb]
            pm.setAttr(name + ".color", convertedRGB)
        else:
            name = prefix + name if name else mat_type
            mat = pm.shadingNode(mat_type, asShader=True, name=name)

        return mat

    @staticmethod
    def assign_mat(objects, mat):
        """Assigns a material to a list of objects or components.

        Parameters:
            objects (list): The objects or components to assign the material to.
            mat (obj): The material to assign.
        """
        try:  # if the mat is a not a known type; try and create the material.
            pm.nodeType(mat)
        except Exception:
            mat = pm.shadingNode(mat, asShader=1)

        shading_group = pm.sets(renderable=True, noSurfaceShader=True, empty=True)
        pm.connectAttr(f"{mat}.outColor", f"{shading_group}.surfaceShader", force=True)

        for obj in pm.ls(objects):
            pm.sets(shading_group, forceElement=obj)

    @classmethod
    def find_by_mat_id(cls, material, objects=None, shell=False):
        """Find objects or faces by the material ID in a 3D scene.

        This function takes as input a material and a set of objects (or entire scene by default) and
        returns a list of objects or faces that are associated with the input material.

        If the 'shell' parameter is set to True, this function returns the complete objects that have
        the input material. If 'shell' is set to False, the function returns the individual faces that
        have the input material.

        Note: If the material is a multi-material (such as VRayMultiSubTex), an error will be raised.

        Args:
            cls: The class object. This argument is implicitly passed and doesn't need to be provided by the user.
            material (str): The material (e.g. 'lambert1') used to find the associated objects/faces.
            objects (list, optional): List of objects (e.g. ['pCube1', 'pCube2']) to be considered.
                                      If not provided, all objects in the scene will be considered.
            shell (bool, optional): Determines whether to return complete objects (True) or individual faces (False).
                                    Default is False.

        Raises:
            TypeError: If material is a multimaterial.

        Returns:
            list: A list of objects or faces (depending on 'shell' parameter) that are associated with the input material.

        """
        if pm.nodeType(material) == "VRayMultiSubTex":
            raise TypeError(
                "Invalid material type. If material is a multimaterial, please select a submaterial."
            )
        # If objects are not specified, consider all objects in the scene
        if not objects:
            objects = pm.ls(geometry=True)  # Get all mesh objects

        # Convert mesh shapes to transform nodes
        objects = node_utils.NodeUtils.get_transform_node(objects)

        # Find the shading groups associated with the material
        shading_groups = pm.listConnections(material, type="shadingEngine")

        objs_with_material = []
        for sg in shading_groups:
            connected_objs = pm.sets(sg, query=True, noIntermediate=True)
            flattened = pm.ls(connected_objs, flatten=True)
            # Only add objects to the list if they are in the specified objects list
            for obj in flattened:
                transform_node = node_utils.NodeUtils.get_transform_node(obj)
                if transform_node in objects:
                    # Check if the object is a face. If not, convert to faces
                    if not isinstance(obj, pm.MeshFace):
                        shape_node = pm.listRelatives(transform_node, shapes=True)[0]
                        face_count = pm.polyEvaluate(shape_node, f=True)
                        faces = [f"{shape_node}.f[{i}]" for i in range(face_count)]
                        objs_with_material.extend(faces)
                    else:
                        objs_with_material.append(obj)

        if shell:
            objs_with_material = set(pm.ls(objs_with_material, objectsOnly=True))

        return objs_with_material

    @staticmethod
    def get_mat_swatch_icon(mat, size=[20, 20]):
        """Get an icon with a color fill matching the given materials RBG value.

        Parameters:
            mat (obj)(str): The material or the material's name.
            size (list): Desired icon size.

        Returns:
            (obj) pixmap icon.
        """
        from PySide2.QtGui import QPixmap, QColor, QIcon

        try:
            # get the string name if a mat object is given.
            matName = mat.name() if not isinstance(mat, (str)) else mat
            # convert from 0-1 to 0-255 value and then to an integer
            r = int(pm.getAttr(matName + ".colorR") * 255)
            g = int(pm.getAttr(matName + ".colorG") * 255)
            b = int(pm.getAttr(matName + ".colorB") * 255)
            pixmap = QPixmap(size[0], size[1])
            pixmap.fill(QColor.fromRgb(r, g, b))

            return QIcon(pixmap)

        except Exception:
            pass

        #  The following is a version that attempts to build a more accurate material swatch icon, but is crashing Maya
        # """Generates a QIcon representing the color swatch of a given Maya material node.

        # Parameters:
        #     mat (str or PyNode): The name or PyNode object of the Maya material. Must be of type 'shadingDependNode'.
        #     size (list of int): A list [width, height] specifying the dimensions of the QIcon in pixels. Default is [20, 20].

        # Returns:
        #     QIcon: A QIcon object filled with the color and transparency attributes of the specified material.
        #            Returns None if the material is invalid or an exception occurs.
        # Example:
        #     icon = get_mat_swatch_icon('lambert1', [20, 20])
        # """
        # from PySide2.QtGui import QPixmap, QPainter, QColor, QIcon

        # try:
        #     # Check if the given 'mat' is a valid Maya material node
        #     if not pm.objectType(mat, isAType="shadingDependNode"):
        #         print(
        #             f"Not a valid material. Expected 'shadingDependNode', got '{mat}' of type '{type(mat).__name__}'."
        #         )
        #         return None

        #     # Initialize QPixmap
        #     pixmap = QPixmap(size[0], size[1])
        #     pixmap.fill(QColor(0, 0, 0, 0))  # Transparent Background

        #     # Initialize QPainter
        #     painter = QPainter(pixmap)

        #     # Fetch material attributes
        #     colorR = pm.getAttr(f"{mat}.colorR") * 255
        #     colorG = pm.getAttr(f"{mat}.colorG") * 255
        #     colorB = pm.getAttr(f"{mat}.colorB") * 255
        #     transparency = pm.getAttr(
        #         f"{mat}.transparencyR"
        #     )  # Assuming R, G, B are the same for simplicity

        #     # Set QColor based on material attributes
        #     brushColor = QColor(colorR, colorG, colorB)
        #     brushColor.setAlpha(255 * (1 - transparency))

        #     # Draw swatch
        #     painter.setBrush(brushColor)
        #     painter.drawRect(0, 0, size[0], size[1])
        #     painter.end()

        #     return QIcon(pixmap)

        # except Exception as e:
        #     print(f"Exception: {e}")
        #     return None


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    ...

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import inspect
import importlib
import pkgutil


__package__ = "mayatk"
__version__ = "0.7.8"

"""Dynamic Attribute Resolver for Module-based Packages

This module implements a dynamic attribute resolver that allows accessing attributes
(classes, methods, and class methods) from a package's modules using the pattern
'<package>.<attribute>'. The resolver builds dictionaries that map attribute names to
their respective module names, enabling efficient and maintainable access to these
attributes.

Key Components:
- CLASS_TO_MODULE: Dictionary mapping class names to their module names.
- METHOD_TO_MODULE: Dictionary mapping method names to their module and class names.
- CLASS_METHOD_TO_MODULE: Dictionary mapping class method names to module and class names.
- build_dictionaries(): Function to populate the dictionaries by inspecting package modules.
- import_module(): Function to import a module by its name.
- get_attribute_from_module(): Function to retrieve an attribute from a module.
- __getattr__(): Special method invoked when an attribute lookup is not found.

Usage:
- Access attributes like classes, methods, and class methods dynamically.
- Improves maintainability by automatically associating attributes with their modules.
- Follows an order of class, method, and class method resolution.

Note: This module is intended for use with package-based projects where attributes are
distributed across modules in a structured manner.
"""

CLASS_TO_MODULE = {}
METHOD_TO_MODULE = {}
CLASS_METHOD_TO_MODULE = {}
IMPORTED_MODULES = {}


def build_dictionaries():
    for importer, modname, ispkg in pkgutil.walk_packages(__path__, __name__ + "."):
        module = importlib.import_module(modname)
        for name, obj in inspect.getmembers(module, inspect.isclass):
            CLASS_TO_MODULE[name] = modname
            for method_name, _ in inspect.getmembers(obj, inspect.isfunction):
                METHOD_TO_MODULE[method_name] = (modname, name)
            for method_name, _ in inspect.getmembers(obj, inspect.ismethod):
                CLASS_METHOD_TO_MODULE[method_name] = (modname, name)


def import_module(module_name):
    if module_name not in IMPORTED_MODULES:
        IMPORTED_MODULES[module_name] = importlib.import_module(module_name)
    return IMPORTED_MODULES[module_name]


def get_attribute_from_module(module, attribute_name, class_name=None):
    if class_name:
        class_obj = getattr(module, class_name)
        return getattr(class_obj, attribute_name)
    return getattr(module, attribute_name)


def __getattr__(name):
    if name in CLASS_TO_MODULE:
        module = import_module(CLASS_TO_MODULE[name])
        return get_attribute_from_module(module, name)

    if name in METHOD_TO_MODULE:
        module_name, class_name = METHOD_TO_MODULE[name]
        module = import_module(module_name)
        return get_attribute_from_module(module, name, class_name)

    if name in CLASS_METHOD_TO_MODULE:
        module_name, class_name = CLASS_METHOD_TO_MODULE[name]
        module = import_module(module_name)
        return get_attribute_from_module(module, name, class_name)

    raise AttributeError(f"module {__package__} has no attribute '{name}'")


# --------------------------------------------------------------------------------------------
# Build dictionaries at the start
build_dictionaries()

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import importlib


def run_test_cases(test_module):
    """Runs all test cases within a test module.

    Parameters:
        test_module (str)(module): A module object or a string representing the module name.
    """
    if isinstance(test_module, str):
        test_module = importlib.import_module(test_module)

    for _, obj in test_module.__dict__.items():
        if isinstance(obj, type) and obj.__module__ == test_module.__name__:
            obj = obj()
            for method_name in dir(obj):
                if method_name.startswith("test_"):
                    test_case = getattr(obj, method_name)
                    test_case()


def run_tests(module_names):
    """Reloads the non-test and test modules and runs the test cases within the test modules.

    Parameters:
        module_names (list): A list of strings representing the names of the non-test modules.
    """
    for module_name in module_names:
        module = __import__(f"mayatk.{module_name}", fromlist=[module_name])
        test_module = __import__(
            f"test.{module_name}_test", fromlist=[f"{module_name}_test"]
        )
        importlib.reload(module)
        importlib.reload(test_module)
        msg_start = f"-> Starting tests for {module_name} .."
        print(f"{'-'*len(msg_start)}\n{msg_start}")
        run_test_cases(test_module)
        print(f"<- {module_name} tests completed.\n{'-'*len(msg_start)}")


# --------------------------------------------------------------------------------------------

if __name__ == "__main__":
    run_tests(
        [
            "core_utils",
            "node_utils",
            "component_utils",
            "edit_utils",
            "xform_utils",
            "rig_utils",
            "mat_utils",
        ]
    )

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)
import pythontk as ptk

# from this package:
from mayatk import core_utils, component_utils


class XformUtils:
    """ """

    @staticmethod
    def move_to(source, target, center=True):
        """Move an object(s) to the given target.

        Parameters:
            source (str/obj/list): The objects to move.
            target (str/obj): The object to move to.
            center (bool): Move to target pivot pos, or the bounding box center of the target.
        """
        if center:  # temporarily move the targets pivot to it's bounding box center.
            orig_target_piv = pm.xform(
                target, q=1, worldSpace=1, rp=1
            )  # get target pivot position.
            pm.xform(target, centerPivots=1)  # center target pivot.
            target_pos = pm.xform(
                target, q=1, worldSpace=1, rp=1
            )  # get the pivot position at center of object.
            pm.xform(
                target, worldSpace=1, rp=orig_target_piv
            )  # return the target pivot to it's orig position.
        else:
            target_pos = pm.xform(
                target, q=1, worldSpace=1, rp=1
            )  # get the pivot position.

        pm.xform(source, translation=target_pos, worldSpace=1, relative=1)

    @staticmethod
    @core_utils.CoreUtils.undo
    def drop_to_grid(
        objects, align="Mid", origin=False, center_pivot=False, freeze_transforms=False
    ):
        """Align objects to Y origin on the grid using a helper plane.

        Parameters:
            objects (str/obj/list): The objects to translate.
            align (bool): Specify which point of the object's bounding box to align with the grid. (valid: 'Max','Mid'(default),'Min')
            origin (bool): Move to world grid's center.
            center_pivot (bool): Center the object's pivot.
            freeze_transforms (bool): Reset the selected transform and all of its children down to the shape level.

        Example:
            drop_to_grid(obj, align='Min') #set the object onto the grid.
        """
        # pm.undoInfo(openChunk=1)
        for obj in pm.ls(objects, transforms=1):
            osPivot = pm.xform(
                obj, q=True, rotatePivot=1, objectSpace=1
            )  # save the object space obj pivot.
            wsPivot = pm.xform(
                obj, q=True, rotatePivot=1, worldSpace=1
            )  # save the world space obj pivot.

            pm.xform(obj, centerPivots=1)  # center pivot
            plane = pm.polyPlane(name="temp#")

            if not origin:
                pm.xform(
                    plane, translation=(wsPivot[0], 0, wsPivot[2]), absolute=1, ws=1
                )  # move the object to the pivot location

            pm.align(obj, plane, atl=1, x="Mid", y=align, z="Mid")
            pm.delete(plane)

            if not center_pivot:
                pm.xform(
                    obj, rotatePivot=osPivot, objectSpace=1
                )  # return pivot to orig position.

            if freeze_transforms:
                pm.makeIdentity(obj, apply=True)
        # pm.undoInfo (closeChunk=1)

    @classmethod
    @core_utils.CoreUtils.undo
    def reset_translation(cls, objects):
        """Reset the translation transformations on the given object(s).

        Parameters:
            objects (str/obj/list): The object(s) to reset the translation values for.
        """
        # pm.undoInfo(openChunk=1)
        for obj in pm.ls(objects):
            pos = pm.objectCenter(obj)  # get the object's current position.
            cls.drop_to_grid(
                obj, origin=1, center_pivot=1
            )  # move to origin and center pivot.
            pm.makeIdentity(obj, apply=1, t=1, r=0, s=0, n=0, pn=1)  # bake transforms
            pm.xform(
                obj, translation=pos
            )  # move the object back to it's original position.
        # pm.undoInfo(closeChunk=1)

    @staticmethod
    def set_translation_to_pivot(node):
        """Set an object’s translation value from its pivot location.

        Parameters:
                node (str/obj/list): An object, or it's name.
        """
        x, y, z = pm.xform(node, query=True, worldSpace=True, rotatePivot=True)
        pm.xform(node, relative=True, translation=[-x, -y, -z])
        pm.makeIdentity(node, apply=True, translate=True)
        pm.xform(node, translation=[x, y, z])

    @staticmethod
    @core_utils.CoreUtils.undo
    def align_pivot_to_selection(align_from=[], align_to=[], translate=True):
        """Align one objects pivot point to another using 3 point align.

        Parameters:
            align_from (list): At minimum; 1 object, 1 Face, 2 Edges, or 3 Vertices.
            align_to (list): The object to align with.
            translate (bool): Move the object with it's pivot.
        """
        # pm.undoInfo(openChunk=1)
        pos = pm.xform(align_to, q=1, translation=True, worldSpace=True)
        center_pos = [  # Get center by averaging of all x,y,z points.
            sum(pos[0::3]) / len(pos[0::3]),
            sum(pos[1::3]) / len(pos[1::3]),
            sum(pos[2::3]) / len(pos[2::3]),
        ]

        vertices = pm.ls(
            pm.polyListComponentConversion(align_to, toVertex=True), flatten=True
        )
        if len(vertices) < 3:
            return

        for obj in pm.ls(align_from, flatten=1):
            plane = pm.polyPlane(
                name="_hptemp#",
                width=1,
                height=1,
                subdivisionsX=1,
                subdivisionsY=1,
                axis=[0, 1, 0],
                createUVs=2,
                constructionHistory=True,
            )[
                0
            ]  # Create and align helper plane.

            pm.select("%s.vtx[0:2]" % plane, vertices[0:3])
            pm.mel.snap3PointsTo3Points(0)

            pm.xform(
                obj,
                rotation=pm.xform(plane, q=True, rotation=True, worldSpace=True),
                worldSpace=True,
            )

            if translate:
                pm.xform(obj, translation=center_pos, worldSpace=True)

            pm.delete(plane)
        # pm.undoInfo(closeChunk=1)

    @staticmethod
    @core_utils.CoreUtils.undo
    def aim_object_at_point(objects, target_pos, aim_vect=(1, 0, 0), up_vect=(0, 1, 0)):
        """Aim the given object(s) at the given world space position.

        Parameters:
            objects (str/obj/list): Transform node(s) of the objects to orient.
            target_pos (obj)(tuple): A point as xyz, or one or more transform nodes at which to aim the other given 'objects'.
            aim_vect (tuple): The vector in local coordinates that points at the target.
            up_vect (tuple): The vector in local coordinates that aligns with the world up vector.

        Example:
            aim_object_at_point(['cube1', 'cube2'], (0, 15, 15))
        """
        if isinstance(target_pos, (tuple, set, list)):
            target = pm.createNode("transform", name="target_helper")
            pm.xform(target, translation=target_pos, absolute=True)
        else:
            target = target_pos  # Assume it's an existing object's name

        for obj in ptk.make_iterable(objects):
            const = pm.aimConstraint(
                target, obj, aim=aim_vect, worldUpVector=up_vect, worldUpType="vector"
            )

        pm.delete(const, target)

    @classmethod
    @core_utils.CoreUtils.undo
    def rotate_axis(cls, objects, target_pos):
        """Aim the given object at the given world space position.
        All rotations in rotated channel, geometry is transformed so
        it does not appear to move during this transformation

        Parameters:
            objects (str/obj/list): Transform node(s) of the objects to orient.
            target_pos (obj)(tuple): A point as xyz, or one or more transform nodes at which to aim the other given 'objects'.
        """
        for obj in pm.ls(objects, type="transform"):
            cls.aim_object_at_point(obj, target_pos)

            try:
                c = obj.verts
            except TypeError:
                c = obj.cp

            wim = pm.getAttr(obj.worldInverseMatrix)
            pm.xform(c, matrix=wim)

            pos = pm.xform(
                obj, q=True, translation=True, absolute=True, worldSpace=True
            )
            pm.xform(c, translation=pos, relative=True, worldSpace=True)

    @staticmethod
    def get_orientation(objects, returned_type="point"):
        """Get an objects orientation as a point or vector.

        Parameters:
            objects (str/obj/list): The object(s) to get the orientation of.
            returned_type (str): The desired returned value type. (valid: 'point'(default), 'vector')

        Returns:
            (tuple)(list) If 'objects' given as a list, a list of tuples will be returned.
        """
        result = []
        for obj in pm.ls(objects, objectsOnly=True):
            world_matrix = pm.xform(obj, q=True, matrix=True, worldSpace=True)
            rAxis = pm.getAttr(obj.rotateAxis)
            if any((rAxis[0], rAxis[1], rAxis[2])):
                print(
                    f"# Warning: {obj} has a modified .rotateAxis of {rAxis} which is included in the result. #"
                )

            if returned_type == "vector":
                from maya.api.OpenMaya import MVector

                ori = (
                    MVector(world_matrix[0:3]),
                    MVector(world_matrix[4:7]),
                    MVector(world_matrix[8:11]),
                )

            else:
                ori = (world_matrix[0:3], world_matrix[4:7], world_matrix[8:11])
            result.append(ori)

        return ptk.format_return(result, objects)

    @staticmethod
    def get_dist_between_two_objects(a, b):
        """Get the magnatude of a vector using the center points of two given objects.

        Parameters:
            a (obj)(str): Object, object name, or point (x,y,z).
            b (obj)(str): Object, object name, or point (x,y,z).

        Returns:
            (float)
        """
        x1, y1, z1 = pm.objectCenter(a)
        x2, y2, z2 = pm.objectCenter(b)

        from math import sqrt

        distance = sqrt(pow((x1 - x2), 2) + pow((y1 - y2), 2) + pow((z1 - z2), 2))

        return distance

    @staticmethod
    def get_center_point(objects):
        """Get the bounding box center point of any given object(s).

        Parameters:
            objects (str)(obj(list): The objects or components to get the center of.

        Returns:
            (tuple) position as xyz float values.
        """
        objects = pm.ls(objects, flatten=True)
        pos = [
            i
            for sublist in [
                pm.xform(s, q=1, translation=1, worldSpace=1, absolute=1)
                for s in objects
            ]
            for i in sublist
        ]
        center_pos = (  # Get center by averaging of all x,y,z points.
            sum(pos[0::3]) / len(pos[0::3]),
            sum(pos[1::3]) / len(pos[1::3]),
            sum(pos[2::3]) / len(pos[2::3]),
        )
        return center_pos

    @classmethod
    def get_bounding_box(cls, objects, value=""):
        """Get information of the given object(s) combined bounding box.

        Parameters:
            objects (str/obj/list): The object(s) or component(s) to query.
                    Multiple objects will be treated as a combined bounding box.
            value (str): The type of value to return. Multiple types can be given
                    separated by '|'. The order given determines the return order.
                    valid (case insensitive): 'xmin', 'xmax', 'ymin', 'ymax', 'zmin', 'zmax', 'size',
                    'x' or 'sizex', 'y' or 'sizey', 'z' or 'sizez', 'volume', 'center' or 'centroid'
        Returns:
            (float)(tuple) Dependant on args.

        Example:
            get_bounding_box(sel, 'center|volume') #returns: [[171.9106216430664, 93.622802734375, -1308.4896240234375], 743.2855185396038]
            get_bounding_box(sel, 'sizeY') #returns: 144.71902465820312
        """
        if "|" in value:  # use recursion to construct the list using each value.
            return tuple(cls.get_bounding_box(objects, i) for i in value.split("|"))

        v = value.lower()
        xmin, ymin, zmin, xmax, ymax, zmax = pm.exactWorldBoundingBox(objects)
        if v == "xmin":
            return xmin
        elif v == "xmax":
            return xmax
        elif v == "ymin":
            return ymin
        elif v == "ymax":
            return ymax
        elif v == "zmin":
            return zmin
        elif v == "zmax":
            return zmax
        elif v == "size":
            return (xmax - xmin, ymax - ymin, zmax - zmin)
        elif v == "sizex" or v == "x":
            return xmax - xmin
        elif v == "sizey" or v == "y":
            return ymax - ymin
        elif v == "sizez" or v == "z":
            return zmax - zmin
        elif v == "minsize":
            return min(xmax - xmin, ymax - ymin, zmax - zmin)
        elif v == "maxsize":
            return max(xmax - xmin, ymax - ymin, zmax - zmin)
        elif v == "volume":
            return (xmax - xmin) * (ymax - ymin) * (zmax - zmin)
        elif v == "center" or v == "centroid":
            return ((xmin + xmax) / 2.0, (ymin + ymax) / 2.0, (zmin + zmax) / 2.0)
        else:
            (xmin, ymin, zmin, xmax, ymax, zmax)

    @classmethod
    def sort_by_bounding_box_value(
        cls, objects, value="volume", descending=True, also_return_value=False
    ):
        """Sort the given objects by their bounding box value.

        Parameters:
            objects (str/obj/list): The objects or components to sort.
            value (str): See 'getBoundingBoxInfo' 'value' parameter.
                            ex. 'xmin', 'xmax', 'sizex', 'volume', 'center' etc.
            descending (bool): Sort the list from the largest value down.
            also_return_value (bool): Instead of just the object; return a
                            list of two element tuples as [(<value>, <obj>)].
        Returns:
            (list)
        """
        valueAndObjs = []
        for obj in pm.ls(objects, flatten=False):
            v = cls.get_bounding_box(obj, value)
            valueAndObjs.append((v, obj))

        sorted_ = sorted(valueAndObjs, key=lambda x: int(x[0]), reverse=descending)
        if also_return_value:
            return sorted_
        return [obj for v, obj in sorted_]

    @staticmethod
    def match_scale(a, b, scale=True, average=False):
        """Scale each of the given objects to the combined bounding box of a second set of objects.

        Parameters:
            a (str/obj/list): The object(s) to get a bounding box size from.
            b (str/obj/list): The object(s) to scale.
            scale (bool): Scale the objects. Else, just return the scale value.
            average (bool): Average the result across all axes.

        Returns:
            (list) scale values as [x,y,z,x,y,z...]
        """
        frm = pm.ls(a, flatten=True)
        to = pm.ls(b, flatten=True)

        xmin, ymin, zmin, xmax, ymax, zmax = pm.exactWorldBoundingBox(frm)
        ax, ay, az = [xmax - xmin, ymax - ymin, zmax - zmin]

        result = []
        for obj in to:
            xmin, ymin, zmin, xmax, ymax, zmax = pm.exactWorldBoundingBox(obj)
            bx, by, bz = [xmax - xmin, ymax - ymin, zmax - zmin]

            oldx, oldy, oldz = pm.xform(obj, q=1, s=1, r=1)

            try:
                diffx, diffy, diffz = [ax / bx, ay / by, az / bz]
            except ZeroDivisionError:
                diffx, diffy, diffz = [1, 1, 1]

            bScaleNew = [oldx * diffx, oldy * diffy, oldz * diffz]

            if average:
                bScaleNew = [sum(bScaleNew) / len(bScaleNew) for _ in range(3)]

            if scale:
                pm.xform(obj, scale=bScaleNew)

            [result.append(i) for i in bScaleNew]

        return result

    @staticmethod
    def align_using_three_points(vertices):
        """Move and align the object defined by the first 3 points to the last 3 points.

        Parameters:
            vertices (list): The first 3 points must be on the same object (i.e. it is the
                    object to be transformed). The second set of points define
                    the position and plane to transform to.
        """
        from maya import OpenMaya

        vertices = pm.ls(vertices, flatten=True)
        objectToMove = pm.ls(vertices[:3], objectsOnly=True)

        p0, p1, p2 = [
            OpenMaya.MPoint(*pm.pointPosition(v, w=True)) for v in vertices[0:3]
        ]
        p3, p4, p5 = [
            OpenMaya.MPoint(*pm.pointPosition(v, w=True)) for v in vertices[3:6]
        ]

        # Translate
        pm.move(*(p3 - p0), objectToMove, r=True, ws=True)

        # First rotation
        axis1 = (p1 - p0).normal()
        axis2 = (p4 - p3).normal()
        # cross_product = axis1 ^ axis2
        angle = axis1.angle(axis2)
        rotation = OpenMaya.MEulerRotation(0, 0, angle).asVector()
        pm.rotate(*rotation, objectToMove, p=p3, r=True, os=True)

        # Second rotation
        axis3 = (p2 - p0).normal()
        axis4 = (p5 - p3).normal()
        # cross_product = axis3 ^ axis4
        angle = axis3.angle(axis4)
        rotation = OpenMaya.MEulerRotation(0, 0, angle).asVector()
        pm.rotate(*rotation, objectToMove, p=p4, r=True, os=True)

    @staticmethod
    def is_overlapping(a, b, tolerance=0.001):
        """Check if the vertices in a and b are overlapping within the given tolerance.

        Parameters:
            a (str/obj): The first object to check. Object can be a component.
            b (str/obj): The second object to check. Object can be a component.
            tolerance (float) = The maximum search distance before a vertex is considered not overlapping.

        Returns:
            (bool)
        """
        vert_setA = pm.ls(pm.polyListComponentConversion(a, toVertex=1), flatten=1)
        vert_setB = pm.ls(pm.polyListComponentConversion(b, toVertex=1), flatten=1)

        closestVerts = component_utils.ComponentUtils.get_closest_verts(
            vert_setA, vert_setB, tolerance=tolerance
        )

        return True if vert_setA and len(closestVerts) == len(vert_setA) else False

    @staticmethod
    def get_vertex_positions(objects, worldSpace=True):
        """Get all vertex positions for the given objects.

        Parameters:
            objects (str/obj/list): The polygon object(s).
            worldSpace (bool): Sample in world or object space.

        Returns:
            (list) Nested lists if multiple objects given.
        """
        import maya.OpenMaya as om

        space = om.MSpace.kWorld if worldSpace else om.MSpace.kObject

        result = []
        for mesh in core_utils.CoreUtils.mfn_mesh_generator(objects):
            points = om.MPointArray()
            mesh.getPoints(points, space)

            result.append(
                [
                    (points[i][0], points[i][1], points[i][2])
                    for i in range(points.length())
                ]
            )
        return ptk.format_return(result, objects)

    @staticmethod
    def hash_points(points, precision=4):
        """Hash the given list of point values.

        Parameters:
            points (list): A list of point values as tuples.
            precision (int): determines the number of decimal places that are retained
                    in the fixed-point representation. For example, with a value of 4, the
                    fixed-point representation would retain 4 decimal place.
        Returns:
            (list) list(s) of hashed tuples.
        """
        nested = ptk.nested_depth(points) > 1
        sets = points if nested else [points]

        def clamp(p):
            return int(p * 10**precision)

        result = []
        for pset in sets:
            result.append([hash(tuple(map(clamp, i))) for i in pset])
        return ptk.format_return(result, nested)

    @classmethod
    def get_matching_verts(cls, a, b, world_space=False):
        """Find any vertices which point locations match between two given mesh.

        Parameters:
            a (str/obj/list): The first polygon object.
            a (str/obj/list): A second polygon object.
            world_space (bool): Sample in world or object space.

        Returns:
            (list) nested tuples with int values representing matching vertex pairs.
        """
        vert_pos_a, vert_pos_b = cls.get_vertex_positions([a, b], world_space)
        hash_a, hash_b = cls.hash_points([vert_pos_a, vert_pos_b])

        matching = set(hash_a).intersection(hash_b)
        return [
            i
            for h in matching
            for i in zip(ptk.indices(hash_a, h), ptk.indices(hash_b, h))
        ]

    @staticmethod
    def order_by_distance(objects, point=[0, 0, 0], reverse=False):
        """Order the given objects by their distance from the given point.

        Parameters:
            objects (str)(int/list): The object(s) to order.
            point (list): A three value float list x, y, z.
            reverse (bool): Reverse the naming order. (Farthest object first)

        Returns:
            (list) ordered objects
        """
        distance = {}
        for obj in pm.ls(objects, flatten=1):
            xmin, ymin, zmin, xmax, ymax, zmax = pm.xform(obj, q=1, boundingBox=1)
            bb_pos = ((xmin + xmax) / 2, (ymin + ymax) / 2, (zmin + zmax) / 2)
            bb_dist = ptk.get_distance(point, bb_pos)

            distance[bb_dist] = obj

        result = [distance[i] for i in sorted(distance)]
        return list(reversed(result)) if reverse else result

    @staticmethod
    @core_utils.CoreUtils.undo
    def align_vertices(mode, average=False, edgeloop=False):
        """Align vertices.

        Parameters:
            mode (int): possible values are align: 0-YZ, 1-XZ, 2-XY, 3-X, 4-Y, 5-Z, 6-XYZ
            average (bool): align to average of all selected vertices. else, align to last selected
            edgeloop (bool): align vertices in edgeloop from a selected edge

        Example:
            align_vertices(mode=3, average=True, edgeloop=True)
        """
        # pm.undoInfo (openChunk=True)
        selectTypeEdge = pm.selectType(query=True, edge=True)

        if edgeloop:
            pm.mel.SelectEdgeLoopSp()  # select edgeloop

        pm.mel.PolySelectConvert(3)  # convert to vertices

        selection = pm.ls(sl=True, flatten=1)
        lastSelected = pm.ls(tail=1, sl=True, flatten=1)
        align_to = pm.xform(lastSelected, q=True, translation=1, worldSpace=1)
        alignX = align_to[0]
        alignY = align_to[1]
        alignZ = align_to[2]

        if average:
            xyz = pm.xform(selection, q=True, translation=1, worldSpace=1)
            x = xyz[0::3]
            y = xyz[1::3]
            z = xyz[2::3]
            alignX = float(sum(x)) / (len(xyz) / 3)
            alignY = float(sum(y)) / (len(xyz) / 3)
            alignZ = float(sum(z)) / (len(xyz) / 3)

        if len(selection) < 2:
            if len(selection) == 0:
                core_utils.CoreUtils.viewport_message("No vertices selected")
            core_utils.CoreUtils.viewport_message(
                "Selection must contain at least two vertices"
            )

        for vertex in selection:
            vertexXYZ = pm.xform(vertex, q=True, translation=1, worldSpace=1)
            vertX = vertexXYZ[0]
            vertY = vertexXYZ[1]
            vertZ = vertexXYZ[2]

            modes = {
                0: (vertX, alignY, alignZ),  # align YZ
                1: (alignX, vertY, alignZ),  # align XZ
                2: (alignX, alignY, vertZ),  # align XY
                3: (alignX, vertY, vertZ),
                4: (vertX, alignY, vertZ),
                5: (vertX, vertY, alignZ),
                6: (alignX, alignY, alignZ),  # align XYZ
            }

            pm.xform(vertex, translation=modes[mode], worldSpace=1)

        if selectTypeEdge:
            pm.selectType(edge=True)
        # pm.undoInfo (closeChunk=True)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    pass

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import os
import unittest
import inspect
import maya.OpenMaya as om
import pymel.core as pm
from mayatk import ComponentUtils


# sfr = pm.melGlobals['cmdScrollFieldReporter']
# pm.cmdScrollFieldReporter(sfr, edit=1, clear=1)


class Main(unittest.TestCase):
    """Main test class."""

    def perform_test(self, cases):
        """Execute the test cases."""
        for case in cases:
            if isinstance(case, str):
                expression = case
                expected_result = cases[case]
                method_name = str(expression).split("(")[0]
            else:
                result, expected_result = case
                method_name = result.__class__.__name__
                expression = None

            try:
                path = os.path.abspath(inspect.getfile(eval(method_name)))
            except (TypeError, IOError):
                path = ""

            if expression:
                result = eval(expression)

            self.assertEqual(
                result,
                expected_result,
                f"\n\n# Error: {path}\n#\tCall: {method_name}({', '.join(map(str, function_args)) if 'function_args' in locals() else ''})\n#\tExpected {type(expected_result)}: {expected_result}\n#\tReturned {type(result)}: {result}",
            )

    @staticmethod
    def replace_mem_address(obj):
        """Replace memory addresses in a string representation of an object with a fixed format of '0x00000000000'.

        Parameters:
                obj (object): The input object. The function first converts this object to a string using the `str` function.

        Returns:
                (str) The string representation of the object with all memory addresses replaced.

        Example:
                >>> replace_mem_address("<class 'str'> <PySide2.QtWidgets.QWidget(0x1ebe2677e80, name='MayaWindow') at 0x000001EBE6D48500>")
                "<class 'str'> <PySide2.QtWidgets.QWidget(0x00000000000, name='MayaWindow') at 0x00000000000>"
        """
        import re

        return re.sub(r"0x[a-fA-F\d]+", "0x00000000000", str(obj))


class CmptUtils_test(Main, ComponentUtils):
    """
    set object mode:
            pm.selectMode(object=1)

    set component mode:
            pm.selectMode(component=1)

    set component mode type:
            pm.selectType(allObjects=1)
            pm.selectType(mc=1)
            pm.selectType(vertex=1)
            pm.selectType(edge=1)
            pm.selectType(facet=1)
            pm.selectType(polymeshUV=1)
            pm.selectType(meshUVShell=1)
    """

    # Tear down the any previous test by creating a new scene:
    pm.mel.file(new=True, force=True)

    # assemble the test scene:
    if not pm.objExists("cyl"):
        cyl = pm.polyCylinder(
            radius=5,
            height=10,
            subdivisionsX=12,
            subdivisionsY=1,
            subdivisionsZ=1,
            name="cyl",
        )

    if not pm.objExists("pln"):
        pln = pm.polyPlane(
            width=20, height=20, subdivisionsX=3, subdivisionsY=3, name="pln"
        )

    if not pm.objExists("sph"):
        sph = pm.polySphere(radius=8, subdivisionsX=6, subdivisionsY=6, name="sph")

    def test_getComponentType(self):
        """ """
        self.perform_test(
            {
                "self.get_component_type('cyl.e[:]')": "e",
                "self.get_component_type('cyl.vtx[:]', 'abv')": "vtx",
                "self.get_component_type('cyl.e[:]', 'int')": 32,
                "self.get_component_type('cyl.e[:]', 'hex')": 0x8000,
            }
        )

    def test_convertAlias(self):
        """ """
        self.perform_test(
            {
                "self.convert_alias('vertex', 'hex')": 0x0001,
                "self.convert_alias(0x0001, 'full')": "Polygon Vertex",
            }
        )

    def test_convertComponentType(self):
        """ """
        self.perform_test(
            {
                "self.convert_component_type('cylShape.vtx[:2]', 'vertex')": [
                    "cylShape.vtx[0:2]"
                ],
                "self.convert_component_type('cylShape.vtx[:2]', 'face')": [
                    "cylShape.f[0:2]",
                    "cylShape.f[11:14]",
                    "cylShape.f[23]",
                ],
                "self.convert_component_type('cylShape.vtx[:2]', 'edge')": [
                    "cylShape.e[0:2]",
                    "cylShape.e[11]",
                    "cylShape.e[24:26]",
                    "cylShape.e[36:38]",
                ],
                "self.convert_component_type('cylShape.vtx[:2]', 'uv')": [
                    "cylShape.map[0:2]",
                    "cylShape.map[12:14]",
                    "cylShape.map[24]",
                ],
            }
        )

    def test_convertIntToComponent(self):
        """ """
        self.perform_test(
            {
                "self.convert_int_to_component('cyl', range(4), 'f')": [
                    "cylShape.f[0:3]"
                ],
                "self.convert_int_to_component('cyl', range(4), 'f', 'int', flatten=True)": [
                    0,
                    1,
                    2,
                    3,
                ],
            }
        )

    def test_filterComponents(self):
        """ """
        self.perform_test(
            {
                "self.filter_components('cyl.vtx[:]', 'cyl.vtx[:2]', 'cyl.vtx[1:23]')": [
                    "cylShape.vtx[0]"
                ],
                "self.filter_components('cyl.f[:]', range(2), range(1, 23))": [
                    "cylShape.f[0]"
                ],
            }
        )

    def test_getComponents(self):
        """ """
        self.assertEqual(
            self.get_components("cyl", "vertex", "str", "", "cyl.vtx[2:23]"),
            [
                "cylShape.vtx[0]",
                "cylShape.vtx[1]",
                "cylShape.vtx[24]",
                "cylShape.vtx[25]",
            ],
        )
        self.assertEqual(
            str(self.get_components("cyl", "vertex", "cyl", "", "cyl.vtx[:23]")),
            "[MeshVertex('cylShape.vtx[24]'), MeshVertex('cylShape.vtx[25]')]",
        )
        self.assertEqual(self.get_components("cyl", "f", "int"), [0, 35])
        self.assertEqual(self.get_components("cyl", "edges"), ["cylShape.e[0:59]"])
        self.assertEqual(
            self.get_components("cyl", "edges", "str", "cyl.e[:2]"),
            [
                "cylShape.e[0]",
                "cylShape.e[1]",
                "cylShape.e[2]",
            ],
        )

    def test_map_components_to_objects(self):
        """Test mapping of components to their respective objects."""
        # Create a sphere
        pm.polySphere(name="mySphere")[0]  # This is the transform node

        actual = self.map_components_to_objects(
            [pm.MeshFace(f"mySphereShape.f[{i}]") for i in range(5)]
        )
        expected = {
            "mySphereShape": [
                pm.MeshFace("mySphereShape.f[0]"),
                pm.MeshFace("mySphereShape.f[1]"),
                pm.MeshFace("mySphereShape.f[2]"),
                pm.MeshFace("mySphereShape.f[3]"),
                pm.MeshFace("mySphereShape.f[4]"),
            ]
        }
        self.assertEqual(actual, expected)

    def test_getContigiousEdges(self):
        """ """
        self.assertEqual(
            self.get_contigious_edges(["cyl.e[:2]"]),
            [{"cylShape.e[1]", "cylShape.e[0]", "cylShape.e[2]"}],
        )
        self.assertEqual(
            self.get_contigious_edges(["cyl.f[0]"]),
            [
                {
                    "cylShape.e[24]",
                    "cylShape.e[0]",
                    "cylShape.e[25]",
                    "cylShape.e[12]",
                }
            ],
        )

    def test_getContigiousIslands(self):
        actual = self.get_contigious_islands("cyl.f[21:26]")
        expected = [
            {"cylShape.f[22]", "cylShape.f[21]", "cylShape.f[23]"},
            {"cylShape.f[26]", "cylShape.f[24]", "cylShape.f[25]"},
        ]

        # Convert each set to a sorted tuple, and sort the lists
        actual_sorted = sorted(tuple(sorted(s)) for s in actual)
        expected_sorted = sorted(tuple(sorted(s)) for s in expected)

        self.assertEqual(actual_sorted, expected_sorted)

    def test_getIslands(self):
        """ """
        if not pm.objExists("cmb"):  # create two objects and combine them.
            pm.polyCylinder(
                radius=5,
                height=10,
                subdivisionsX=5,
                subdivisionsY=1,
                subdivisionsZ=1,
                name="cmbA",
            )
            pm.polyCylinder(
                radius=5,
                height=6,
                subdivisionsX=5,
                subdivisionsY=1,
                subdivisionsZ=1,
                name="cmbB",
            )
            # polyUnite returns a tuple of the resulting object and the node that performed the operation.
            cmb, _ = pm.polyUnite(
                "cmbA", "cmbB", ch=1, mergeUVSets=1, centerPivot=1, name="cmb"
            )

        # get_islands returns a generator so convert to list.
        self.assertEqual(
            list(self.get_islands(cmb)),
            [
                [
                    "cmb.f[0]",
                    "cmb.f[5]",
                    "cmb.f[4]",
                    "cmb.f[9]",
                    "cmb.f[6]",
                    "cmb.f[1]",
                    "cmb.f[10]",
                    "cmb.f[11]",
                    "cmb.f[14]",
                    "cmb.f[8]",
                    "cmb.f[7]",
                    "cmb.f[3]",
                    "cmb.f[13]",
                    "cmb.f[2]",
                    "cmb.f[12]",
                ],
                [
                    "cmb.f[15]",
                    "cmb.f[20]",
                    "cmb.f[19]",
                    "cmb.f[24]",
                    "cmb.f[21]",
                    "cmb.f[16]",
                    "cmb.f[25]",
                    "cmb.f[26]",
                    "cmb.f[29]",
                    "cmb.f[23]",
                    "cmb.f[22]",
                    "cmb.f[18]",
                    "cmb.f[28]",
                    "cmb.f[17]",
                    "cmb.f[27]",
                ],
            ],
        )

    def test_getBorderComponents(self):
        """ """
        self.perform_test(
            {
                "self.get_border_components('pln', 'vtx')": [
                    "plnShape.vtx[0:4]",
                    "plnShape.vtx[7:8]",
                    "plnShape.vtx[11:15]",
                ],
                "self.get_border_components('pln', 'face')": [
                    "plnShape.f[0:3]",
                    "plnShape.f[5:8]",
                ],
                "self.get_border_components('pln')": [
                    "plnShape.e[0:2]",
                    "plnShape.e[4]",
                    "plnShape.e[6]",
                    "plnShape.e[8]",
                    "plnShape.e[13]",
                    "plnShape.e[15]",
                    "plnShape.e[20:23]",
                ],
                "self.get_border_components('pln.e[:]')": [
                    "plnShape.e[0:2]",
                    "plnShape.e[4]",
                    "plnShape.e[6]",
                    "plnShape.e[8]",
                    "plnShape.e[13]",
                    "plnShape.e[15]",
                    "plnShape.e[20:23]",
                ],
                "self.get_border_components(['pln.e[9]','pln.e[10]', 'pln.e[12]', 'pln.e[16]'], 'f', component_border=True)": [
                    "plnShape.f[1]",
                    "plnShape.f[3:5]",
                    "plnShape.f[7]",
                ],
                "self.get_border_components('pln.f[3:4]', 'f', component_border=True)": [
                    "plnShape.f[0:1]",
                    "plnShape.f[5:7]",
                ],
                "self.get_border_components('pln.f[3:4]', 'vtx', component_border=True)": [
                    "plnShape.vtx[4:6]",
                    "plnShape.vtx[8:10]",
                ],
                "self.get_border_components('pln.vtx[6]', 'e', component_border=True)": [
                    "plnShape.e[5]",
                    "plnShape.e[9]",
                    "plnShape.e[11:12]",
                ],
            }
        )

    def test_getClosestVerts(self):
        """ """
        self.perform_test(
            {
                "self.get_closest_verts('pln.vtx[:10]', 'pln.vtx[11:]', 6.667)": [
                    ("plnShape.vtx[7]", "plnShape.vtx[11]"),
                    ("plnShape.vtx[8]", "plnShape.vtx[12]"),
                    ("plnShape.vtx[9]", "plnShape.vtx[13]"),
                    ("plnShape.vtx[10]", "plnShape.vtx[11]"),
                    ("plnShape.vtx[10]", "plnShape.vtx[14]"),
                ],
            }
        )

    def test_getClosestVertex(self):
        """ """
        self.perform_test(
            {
                "self.get_closest_vertex('plnShape.vtx[0]', 'cyl', returned_type='int')": {
                    "plnShape.vtx[0]": 6
                },
                "self.get_closest_vertex('plnShape.vtx[0]', 'cyl')": {
                    "plnShape.vtx[0]": "cylShape.vtx[6]"
                },
                "self.get_closest_vertex('plnShape.vtx[2:3]', 'cyl')": {
                    "plnShape.vtx[2]": "cylShape.vtx[9]",
                    "plnShape.vtx[3]": "cylShape.vtx[9]",
                },
            }
        )

    def test_getEdgePath(self):
        """ """
        self.perform_test(
            {
                "self.get_edge_path('sph.e[12]', 'edgeLoop')": [
                    "sphShape.e[12]",
                    "sphShape.e[17]",
                    "sphShape.e[16]",
                    "sphShape.e[15]",
                    "sphShape.e[14]",
                    "sphShape.e[13]",
                ],
                "self.get_edge_path('sph.e[12]', 'edgeLoop', 'int')": [
                    12,
                    17,
                    16,
                    15,
                    14,
                    13,
                ],
                "self.get_edge_path('sph.e[12]', 'edgeRing')": [
                    "sphShape.e[0]",
                    "sphShape.e[6]",
                    "sphShape.e[12]",
                    "sphShape.e[18]",
                    "sphShape.e[24]",
                ],
                "self.get_edge_path(['sph.e[43]', 'sph.e[46]'], 'edgeRingPath')": [
                    "sphShape.e[43]",
                    "sphShape.e[42]",
                    "sphShape.e[47]",
                    "sphShape.e[46]",
                ],
                "self.get_edge_path(['sph.e[54]', 'sph.e[60]'], 'edgeLoopPath')": [
                    "sphShape.e[60]",
                    "sphShape.e[48]",
                    "sphShape.e[42]",
                    "sphShape.e[36]",
                    "sphShape.e[30]",
                    "sphShape.e[54]",
                ],
            }
        )

    def test_get_normal(self):
        """ """
        face = pm.PyNode("cylShape.f[0]")
        expected_normal = om.MVector(0.7071068109888711, 0.0, -0.7071067513842227)
        result_normal = self.get_normal(face)
        self.assertAlmostEqual(result_normal.x, expected_normal.x)
        self.assertAlmostEqual(result_normal.y, expected_normal.y)
        self.assertAlmostEqual(result_normal.z, expected_normal.z)

    def test_get_normal_vector(self):
        """ """
        # Create a cube and get its first face
        cube = pm.polyCube()[0]
        face = cube.f[0]

        # Retrieve the normal directly from Maya for the first face of the cube
        expected_normal = pm.polyInfo(face, fn=True)[0].split()[2:5]
        expected_normal = [float(i) for i in expected_normal]

        # Get the normal using the get_normal_vector function
        result_normal = self.get_normal_vector(face)

        # Since result_normal is a dictionary, we retrieve the value using the face's ID as key
        self.assertAlmostEqual(result_normal[0][0], expected_normal[0])
        self.assertAlmostEqual(result_normal[0][1], expected_normal[1])
        self.assertAlmostEqual(result_normal[0][2], expected_normal[2])

    def test_get_normal_angle(self):
        """ """
        # Create a cube and get its first two edges
        cube = pm.polyCube()[0]
        edge1 = cube.e[0]
        edge2 = cube.e[1]

        # The angle between the normals of the faces connected by the first edge is 90 degrees
        # The angle between the normals of the faces connected by the second edge is also 90 degrees
        self.assertEqual(self.get_normal_angle(edge1), 90.0)
        self.assertEqual(self.get_normal_angle(edge2), 90.0)

    def test_average_normals(self):
        """ """
        # Create a cube
        cube = pm.polyCube()[0]

        # Average the normals of the cube
        self.average_normals([cube])

        # The average normal for a cube should be equal for each face
        normals = self.get_normal_vector(cube.f)

        # Expected normals for a cube
        expected_normals = {
            0: [0.0, 0.0, 1.0],
            1: [0.0, 1.0, 0.0],
            2: [0.0, 0.0, -1.0],
            3: [0.0, -1.0, 0.0],
            4: [1.0, 0.0, 0.0],
            5: [-1.0, 0.0, 0.0],
        }

        # Verify that the normals are approximately equal to the expected normals
        for key, normal in normals.items():
            self.assertAlmostEqual(normal[0], expected_normals[key][0])
            self.assertAlmostEqual(normal[1], expected_normals[key][1])
            self.assertAlmostEqual(normal[2], expected_normals[key][2])

    def test_get_edges_by_normal_angle(self):
        """ """
        self.assertEqual(
            self.get_edges_by_normal_angle("cyl", 50, 130),
            [pm.PyNode("cylShape.e[{}]".format(i)) for i in range(24)],
        )

    def test_set_edge_hardness(self):
        """Test the set_edge_hardness method"""
        # Create a cube
        cube = pm.polyCube()[0]

        # Apply different hardness to different edges of the cube
        self.set_edge_hardness([cube], 45, upper_hardness=60, lower_hardness=20)

        # ** temp disabled: polySoftEdge is throwing an error on edge angle query. **
        # The first edge of the cube should have a hardness of 60 (upper hardness)
        # self.assertEqual(pm.polySoftEdge(cube.e[0], q=True, angle=True), 60.0)

        # If we now set the threshold to a higher value and apply again
        self.set_edge_hardness([cube], 135, upper_hardness=80, lower_hardness=30)

        # The first edge of the cube should now have a hardness of 30 (lower hardness)
        # self.assertEqual(pm.polySoftEdge(cube.e[0], q=True, angle=True), 30.0)

    def test_get_faces_with_similar_normals(self):
        """ """
        # Create a cube and sphere
        cube = pm.polyCube()[0]
        sphere = pm.polySphere()[0]

        # Get faces of the cube and sphere
        cube_faces = cube.f[:]
        # sphere_faces = sphere.f[:]

        # For a cube and a sphere, there should be faces with similar normals.
        # Since we do not know which faces will have similar normals,
        # we cannot predict the exact number of similar faces.
        similar_faces = self.get_faces_with_similar_normals(
            cube_faces, transforms=[sphere], range_x=0.1, range_y=0.1, range_z=0.1
        )

        # So we just assert that there are some similar faces found
        self.assertTrue(len(similar_faces) > 0)

    def test_transfer_normals(self):
        """ """
        # Create two cubes
        source = pm.polyCube()[0]
        target = pm.polyCube()[0]

        # Get the vertices of the first face of the source cube
        vertices = pm.polyListComponentConversion(
            source.f[0], fromFace=True, toVertex=True
        )
        vertices = pm.ls(vertices, flatten=True)

        # Change the normal of the vertices
        for vertex in vertices:
            pm.polyNormalPerVertex(vertex, normalXYZ=[1.0, 0.0, 0.0])

        # Transfer normals from source to target
        self.transfer_normals(source, target)

        # The normal for the first face of the target should now match the source
        source_normal = self.get_normal_vector(source.f[0])[0]
        target_normal = self.get_normal_vector(target.f[0])[0]

        self.assertAlmostEqual(source_normal[0], target_normal[0])
        self.assertAlmostEqual(source_normal[1], target_normal[1])
        self.assertAlmostEqual(source_normal[2], target_normal[2])

    def test_getComponentsByNumberOfConnected(self):
        """ """
        self.perform_test(
            {
                "self.filter_components_by_connection_count(['sph.f[18:23]', 'sph.f[30:35]'], 3, 'e')": [
                    "sphShape.f[30]",
                    "sphShape.f[31]",
                    "sphShape.f[32]",
                    "sphShape.f[33]",
                    "sphShape.f[34]",
                    "sphShape.f[35]",
                ],
                "self.filter_components_by_connection_count('pln.vtx[:]', (0,2), 'e')": [
                    "plnShape.vtx[0]",
                    "plnShape.vtx[3]",
                    "plnShape.vtx[12]",
                    "plnShape.vtx[15]",
                ],
            }
        )

    def test_getVertexNormal(self):
        """ """
        import maya.api.OpenMaya as om

        self.perform_test(
            {
                "self.get_vertex_normal('pln.vtx[2]', angle_weighted=False)": om.MVector(
                    0, 1, 0
                ),
            }
        )

    def test_getVectorFromComponents(self):
        """ """
        self.perform_test(
            {
                "self.get_vector_from_components('pln.f[:]')": (0.0, 1.0, 0.0),
                "self.get_vector_from_components(['cyl.f[7]', 'cyl.f[8]'])": (
                    0.0,
                    0.0,
                    0.43982641180356435,
                ),
            }
        )

    def test_orient_shells(self):
        """ """
        # Create a plane with a UV map
        plane = pm.polyPlane()[0]
        pm.polyUVSet(plane, create=True, uvSet="map1")

        # Orient the UV shells of the plane
        self.orient_shells([plane])

        # Check if UV shells are oriented as expected
        uv_shells = pm.polyUVSet(plane, query=True, allUVSets=True)
        self.assertTrue(uv_shells)

    def test_move_to_uv_space(self):
        """Test the move_to_uv_space method."""
        # Create a plane
        plane = pm.polyPlane()[0]

        # Move the UVs of the plane
        self.move_to_uv_space([plane], 1, 1, relative=False)

        # Get the UVs of the plane
        uvs = pm.polyListComponentConversion(plane.f, fromFace=True, toUV=True)
        uvs = pm.ls(uvs, flatten=True)

        for uv in uvs:
            # Query the UV coordinate
            u, v = pm.polyEditUV(uv, query=True)

            # Check that the UV coordinate is approximately equal to [1, 1]
            self.assertAlmostEqual(u, 1, places=5)
            self.assertAlmostEqual(v, 1, places=5)

    def test_get_uv_shell_sets(self):
        """ """
        # Create a plane with a UV map
        plane = pm.polyPlane()[0]
        pm.polyUVSet(plane, create=True, uvSet="map1")

        # Get the UV shell sets of the plane
        uv_shell_sets = self.get_uv_shell_sets([plane])

        # Check if the UV shell sets are as expected
        self.assertTrue(uv_shell_sets)  # Should not be empty

    def test_get_uv_shell_border_edges(self):
        """ """
        # Create a plane with a UV map
        plane = pm.polyPlane()[0]
        pm.polyUVSet(plane, create=True, uvSet="map1")

        # Get the UV shell border edges of the plane
        uv_shell_border_edges = self.get_uv_shell_border_edges([plane])

        # Check if the UV shell border edges are as expected
        self.assertTrue(uv_shell_border_edges)  # Should not be empty

    def test_transfer_uvs(self):
        """ """
        # Create two planes
        source = pm.polyPlane()[0]
        target = pm.polyPlane()[0]

        # Set a UV for the first vertex of the first face of the source plane
        source_uv_set = "map1"
        target_uv_set = "map2"
        pm.polyUVSet(source, create=True, uvSet=source_uv_set)
        pm.polyUVSet(target, create=True, uvSet=target_uv_set)

        # Get the UVs of the first face of the source plane
        source_uvs = pm.polyListComponentConversion(
            source.f[0], fromFace=True, toUV=True
        )
        source_uvs = pm.ls(source_uvs, flatten=True)

        # Edit the first UV of the source plane
        pm.polyEditUV(source_uvs[0], u=0.5, v=0.5)

        # Transfer UVs from source to target
        self.transfer_uvs(source, target)

        # Get the UVs of the first face of the target plane
        target_uvs = pm.polyListComponentConversion(
            target.f[0], fromFace=True, toUV=True
        )
        target_uvs = pm.ls(target_uvs, flatten=True)

        # The UV for the first vertex of the first face of the target should now match the source
        source_u, source_v = pm.polyEditUV(source_uvs[0], query=True)
        target_u, target_v = pm.polyEditUV(target_uvs[0], query=True)

        self.assertAlmostEqual(source_u, target_u)
        self.assertAlmostEqual(source_v, target_v)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main(exit=False)


# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

# """
# def test_(self):
#   '''
#   '''
#   self.perform_test({
#       # "self.": '',
#   })


# def test_(self):
#   '''
#   '''
#   self.perform_test({
#       # "self.": '',
#   })


# def test_(self):
#   '''
#   '''
#   self.perform_test({
#       # "self.": '',
#   })


# def test_(self):
#   '''
#   '''
#   self.perform_test({
#       # "self.": '',
#   })
# """

# Deprecated ---------------------



# !/usr/bin/python
# coding=utf-8
import unittest
import importlib
import pymel.core as pm
from mayatk import mat_utils

importlib.reload(mat_utils)


class MatUtilsTest(unittest.TestCase, mat_utils.MatUtils):
    def test_get_mats(self):
        """Test the get_mats method."""
        # Create a sphere and assign a lambert material to it
        sphere = pm.polySphere()[0]
        lambert = pm.shadingNode("lambert", asShader=True)
        pm.select(sphere)
        pm.hyperShade(assign=lambert)

        # Get the materials of the sphere
        mats = self.get_mats(sphere)

        # Print materials for debug
        print(f"Materials of the sphere: {mats}")

        # Check if the lambert material is in the materials of the sphere
        self.assertIn(lambert, mats)

        # Get the materials of a face of the sphere
        face = sphere.f[0]
        face_mats = self.get_mats(face)

        # Print materials for debug
        print(f"Materials of the face: {face_mats}")

        # Check if the lambert material is in the materials of the face
        self.assertIn(lambert, face_mats)

    def test_get_scene_mats(self):
        """Test the get_scene_mats method."""
        # Create a lambert material
        lambert = pm.shadingNode("lambert", asShader=True)

        # Get all materials in the scene
        scene_mats = self.get_scene_mats()

        # Check if the lambert material is in the scene materials
        self.assertIn(lambert, scene_mats)

    def test_get_fav_mats(self):
        """Test the get_fav_mats method."""
        # Get the favorite materials
        fav_mats = self.get_fav_mats()

        # Check if the favorite materials are not None
        self.assertIsNotNone(fav_mats)

    def test_create_mat(self):
        """Test the create_material method."""
        # Create a random material
        random_mat = self.create_mat(mat_type="random")

        # Check if the random material is not None
        self.assertIsNotNone(random_mat)

        # Optionally, you can test other material types
        lambert_mat = self.create_mat(mat_type="lambert")
        self.assertIsNotNone(lambert_mat)

    def test_assign_mat(self):
        """Test the assign_mat method."""
        # Create a sphere
        sphere = pm.polySphere()[0]

        # Create a lambert material
        lambert = pm.shadingNode("lambert", asShader=True)

        # Assign the lambert material to the sphere
        mat_utils.MatUtils.assign_mat([sphere], lambert)

        # Print the materials of the sphere after assignment
        print(
            "Materials after assignment:",
            pm.ls(pm.listConnections(sphere), materials=True),
        )

        # Get the materials of the sphere
        mats = list(mat_utils.MatUtils.get_mats(sphere))

        # Print the materials retrieved by get_mats
        print("Materials from get_mats:", mats)

        # Check that the lambert material is in the materials of the sphere
        self.assertIn(lambert, mats)

    def test_find_by_mat_id(self):
        cube = pm.polyCube()[0]
        lambert = pm.shadingNode("lambert", asShader=True)

        # # Test that find_by_mat_id raises error with multi-material
        # multi_mat = pm.createNode("VRayMultiSubTex")
        # with self.assertRaises(TypeError):
        #     self.find_by_mat_id(multi_mat, objects=[cube])

        # Assign lambert material to face 0 of the cube
        self.assign_mat([f"{cube.getShape().name()}.f[0]"], lambert)

        # Test that find_by_mat_id finds the face when no objects are specified
        objs = self.find_by_mat_id(lambert)
        objs_as_strings = [obj.name() for obj in objs]
        self.assertIn(f"{cube.getShape().name()}.f[0]", objs_as_strings)

        # Test that find_by_mat_id finds the face from the given objects
        objs = self.find_by_mat_id(lambert, objects=[cube])
        objs_as_strings = [obj.name() for obj in objs]
        self.assertIn(f"{cube.getShape().name()}.f[0]", objs_as_strings)

        # Test that find_by_mat_id returns complete objects when shell is True
        objs = self.find_by_mat_id(lambert, shell=True)
        self.assertIn(cube.getShape(), objs)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main(exit=False)


# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)
import pythontk as ptk

# from this package:
from mayatk.core_utils import CoreUtils, Preview


class DuplicateLinear:
    @staticmethod
    def duplicate_linear(
        objects,
        num_copies,
        translate=(0, 0, 0),
        rotate=(0, 0, 0),
        scale=(1, 1, 1),
        weight_bias=0.5,
        weight_curve=4,
    ):
        weight_factor = (
            2 * (weight_bias - 0.5) if weight_bias >= 0.5 else 2 * (0.5 - weight_bias)
        )

        originals_to_copies = {}

        for node in objects:
            # Get the parent of the original object
            parent_node = node.getParent()
            copies = []  # List to hold copies of current node
            original_scale = node.scale.get()
            scale_steps = [
                (final_scale - init_scale)
                for final_scale, init_scale in zip(scale, original_scale)
            ]

            for i in range(num_copies):
                dup = pm.instance(node)[0]

                # Set the parent of the copy to the same as the original object
                pm.parent(dup, parent_node)

                x = (i + 1) / num_copies  # Adjusted line
                f_x = ptk.lerp(
                    x,
                    x**weight_curve
                    if weight_bias >= 0.5
                    else 1 - (1 - x) ** weight_curve,
                    weight_factor,
                )

                # Now transformations are adjusted directly by f_x
                dup.translate.set(dup.translate.get() + (pm.dt.Vector(translate) * f_x))
                dup.rotate.set(dup.rotate.get() + (pm.dt.Vector(rotate) * f_x))
                dup.scale.set(
                    [
                        init_scale + (b * f_x)
                        for init_scale, b in zip(original_scale, scale_steps)
                    ]
                )

                copies.append(dup)

            originals_to_copies[node] = copies

        return originals_to_copies


class DuplicateLinearSlots:
    def __init__(self):
        self.sb = self.switchboard()
        self.ui = self.sb.duplicate_linear

        self.preview = Preview(
            self.ui.chk000,
            self.ui.b000,
            operation_func=self.perform_duplicate_linear,
            message_func=self.sb.message_box,
        )
        self.sb.connect_multi(
            self.ui,
            "s000-11",
            "valueChanged",
            self.preview.refresh,
        )

    def perform_duplicate_linear(self):
        """Perform the linear duplication operation."""
        objects = pm.ls(sl=True, type="transform")
        num_copies = self.ui.s009.value()
        translate = (
            self.ui.s000.value(),
            self.ui.s001.value(),
            self.ui.s002.value(),
        )
        rotate = (
            self.ui.s003.value(),
            self.ui.s004.value(),
            self.ui.s005.value(),
        )
        scale = (
            self.ui.s006.value(),
            self.ui.s007.value(),
            self.ui.s008.value(),
        )
        weight_bias = self.ui.s010.value()
        weight_curve = self.ui.s011.value()

        self.copies = DuplicateLinear.duplicate_linear(
            objects,
            num_copies,
            translate,
            rotate,
            scale,
            weight_bias,
            weight_curve,
        )


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import os
    from uitk import Switchboard

    parent = CoreUtils.get_main_window()
    ui_file = os.path.join(os.path.dirname(__file__), "duplicate_linear.ui")
    sb = Switchboard(parent, ui_location=ui_file, slot_location=DuplicateLinearSlots)

    sb.current_ui.set_attributes(WA_TranslucentBackground=True)
    sb.current_ui.set_flags(
        Tool=True, FramelessWindowHint=True, WindowStaysOnTopHint=True
    )
    sb.current_ui.set_style(theme="dark", style_class="translucentBgWithBorder")

    sb.current_ui.show(pos="screen", app_exec=True)

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)


class Preview:
    """A class to handle preview functionality in a GUI.
    It provides functionalities for preview toggle, enabling/disabling preview,
    refreshing the preview, and finalizing the changes.

    Example:
        from PySide2.QtWidgets import QApplication, QPushButton, QCheckBox, QSlider

        def operation_func():
            print("Operation performed.")

        def finalize_func():
            print("Changes finalized.")

        app = QApplication([])  # Required to create a QApplication instance first

        preview_checkbox = QCheckBox()  # Replace with actual instance.
        create_button = QPushButton()  # Replace with actual instance.
        value_slider = QSlider()  # Example widget that emits a signal

        preview = Preview(preview_checkbox, create_button, operation_func, finalize_func)

        # Connect value_slider's valueChanged signal to the preview's refresh method
        value_slider.valueChanged.connect(preview.refresh)

        # Checking the checkbox will execute `operation_func` and print "Operation performed."
        preview_checkbox.setChecked(True)

        # Moving the slider will also call `operation_func`
        value_slider.setValue(5)

        # Pressing the button will execute `finalize_func` and print "Changes finalized."
        create_button.click()

        app.exec_()  # Start the application's event loop
    """

    def __init__(self, *args, **kwargs):
        """Initialize the Preview class with preview-related UI elements and functions."""
        self.init_preview_mixin(*args, **kwargs)

    def init_preview_mixin(
        self,
        preview_checkbox,
        create_button,
        operation_func=None,
        finalize_func=None,
        message_func=print,
    ):
        """Initialize the mixin for the preview. Connects UI elements to their respective functionalities.

        Parameters:
            preview_checkbox (QWidget): The UI element for the preview checkbox.
            create_button (QWidget): The UI element for the create/apply button.
            operation_func (callable, optional): The function to execute for the preview operation.
            finalize_func (callable, optional): The function to finalize the changes.
            message_func (callable, optional): The function to display messages. Defaults to print.
        """
        self.preview_checkbox = preview_checkbox
        self.create_button = create_button
        self.operation_func = operation_func
        self.finalize_func = finalize_func
        self.message_func = message_func
        self.needs_undo = False

        self.preview_checkbox.clicked.connect(self.toggle)
        self.create_button.clicked.connect(self.finalize_changes)
        self.window = self.create_button.window()
        if hasattr(self.window, "on_hide"):
            # Un-check the preview button on window hide event.
            self.window.on_hide.connect(self.disable)

    def toggle(self, state):
        """Toggle the preview state based on the provided state.

        Parameters:
            state (bool): The state to set the preview to. True for enabled, False for disabled.
        """
        if state:
            self.enable()
        else:
            self.disable()

    def enable(self):
        """Enable the preview, perform the operation, and enable the 'Apply Changes' button."""
        if pm.selected():
            self.needs_undo = False
            self.refresh()
            self.preview_checkbox.setChecked(True)
            self.create_button.setEnabled(True)
            # Mute the command reporting.
            pm.commandEcho(state=False)
        else:
            self.message_func("No objects selected.")
            self.disable()

    def disable(self):
        """Disable the preview, undo the last operation if needed, and disable the 'Apply Changes' button."""
        self.undo_if_needed()
        self.preview_checkbox.setChecked(False)
        self.create_button.setEnabled(False)
        # Restore the command reporting to its default state.
        pm.commandEcho(state=True)

    def refresh(self, *args):
        """Refresh the preview if the checkbox is checked and there's no previous operation pending.

        Note:
            *args is needed because the method may be called as a slot in PyQt,
            which might pass extra arguments depending on the signal that triggered the slot.
        """
        if self.operation_func is None:
            raise ValueError("Operation function not defined!")

        if not self.preview_checkbox.isChecked():
            return

        # If we've already performed an operation, undo it before doing a new one.
        self.undo_if_needed()

        pm.undoInfo(openChunk=True)
        self.operation_func()
        pm.undoInfo(closeChunk=True)

        self.needs_undo = True

    def finalize_changes(self):
        """Apply Changes and emit signal.
        This disables the preview and, if defined, calls the finalize function.
        """
        self.needs_undo = False
        self.disable()

        if self.finalize_func:
            self.finalize_func()

    def undo_if_needed(self):
        """Undo the previous operation if there is one pending.

        This method will attempt to perform an undo operation if `self.needs_undo` is True.
        If there's no command to undo, it catches the RuntimeError and passes.
        """
        if self.needs_undo:
            try:
                pm.undo()
            except RuntimeError:  # There are no more commands to undo.
                pass


# -----------------------------------------------------------------------------


if __name__ == "__main__":
    pass

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import os
import unittest
import inspect
import pymel.core as pm
from mayatk import EditUtils


# sfr = pm.melGlobals['cmdScrollFieldReporter']
# pm.cmdScrollFieldReporter(sfr, edit=1, clear=1)


class Main(unittest.TestCase):
    """Main test class."""

    def perform_test(self, cases):
        """Execute the test cases."""
        for case in cases:
            if isinstance(case, str):
                expression = case
                expected_result = cases[case]
                method_name = str(expression).split("(")[0]
            else:
                result, expected_result = case
                method_name = result.__class__.__name__
                expression = None

            try:
                path = os.path.abspath(inspect.getfile(eval(method_name)))
            except (TypeError, IOError):
                path = ""

            if expression:
                result = eval(expression)

            self.assertEqual(
                result,
                expected_result,
                f"\n\n# Error: {path}\n#\tCall: {method_name}({', '.join(map(str, function_args)) if 'function_args' in locals() else ''})\n#\tExpected {type(expected_result)}: {expected_result}\n#\tReturned {type(result)}: {result}",
            )

    @staticmethod
    def replace_mem_address(obj):
        """Replace memory addresses in a string representation of an object with a fixed format of '0x00000000000'.

        Parameters:
                obj (object): The input object. The function first converts this object to a string using the `str` function.

        Returns:
                (str) The string representation of the object with all memory addresses replaced.

        Example:
                >>> replace_mem_address("<class 'str'> <PySide2.QtWidgets.QWidget(0x1ebe2677e80, name='MayaWindow') at 0x000001EBE6D48500>")
                "<class 'str'> <PySide2.QtWidgets.QWidget(0x00000000000, name='MayaWindow') at 0x00000000000>"
        """
        import re

        return re.sub(r"0x[a-fA-F\d]+", "0x00000000000", str(obj))


class EditUtils_test(Main, EditUtils):
    """ """

    # Tear down the any previous test by creating a new scene:
    pm.mel.file(new=True, force=True)

    # assemble the test scene:
    if not pm.objExists("cube1"):
        cube1 = pm.polyCube(
            width=5,
            height=5,
            depth=5,
            subdivisionsX=1,
            subdivisionsY=1,
            subdivisionsZ=1,
            name="cube1",
        )

    if not pm.objExists("cube2"):
        cube2 = pm.polyCube(
            width=2,
            height=4,
            depth=8,
            subdivisionsX=3,
            subdivisionsY=3,
            subdivisionsZ=3,
            name="cube2",
        )

    if not pm.objExists("cyl"):
        cyl = pm.polyCylinder(
            radius=5,
            height=10,
            subdivisionsX=12,
            subdivisionsY=1,
            subdivisionsZ=1,
            name="cyl",
        )

    def test_rename(self):
        """ """
        self.perform_test(
            {
                "self.rename('cube1', 'newName')": None,
                "self.rename('newName', 'cube1')": None,
            }
        )

    def test_setCase(self):
        """ """
        self.perform_test(
            {
                "self.set_case('cube1', 'lower')": None,
            }
        )

    def test_setSuffixByObjLocation(self):
        """ """
        if not pm.objExists("c1"):
            c1 = pm.polyCube(
                width=2,
                height=2,
                depth=8,
                subdivisionsX=1,
                subdivisionsY=1,
                subdivisionsZ=1,
                name="c1",
            )

        if not pm.objExists("c2"):
            c2 = pm.polyCube(
                width=8,
                height=2,
                depth=2,
                subdivisionsX=1,
                subdivisionsY=1,
                subdivisionsZ=1,
                name="c2",
            )
            pm.move(0, 0, 5, c2)

        self.perform_test(
            {
                "self.append_location_based_suffix(['c1', 'c2'])": None,
            }
        )

    def test_snapClosestVerts(self):
        """ """
        self.perform_test(
            {
                "self.snap_closest_verts('cube1', 'cube2')": None,
            }
        )

    def test_mergeVertices(self):
        """ """
        self.perform_test(
            {
                "self.merge_vertices('cube1')": None,
            }
        )

    def test_deleteAlongAxis(self):
        """ """
        self.perform_test(
            {
                "self.delete_along_axis('cube1')": None,
            }
        )

    def test_getAllFacesOnAxis(self):
        """ """
        self.perform_test(
            {
                "self.get_all_faces_on_axis('cube1')": [],  # faces should have been deleted by the previous test 'delete_along_axis'.
            }
        )

    def test_cleanGeometry(self):
        """ """
        self.perform_test(
            {
                "self.clean_geometry('cyl')": None,
            }
        )

    def test_getOverlappingDupObjects(self):
        """ """
        self.perform_test(
            {
                "self.get_overlapping_duplicates(['cyl', 'cube1', 'cube2'])": set(),
            }
        )

    def test_findNonManifoldVertex(self):
        """ """
        self.perform_test(
            {
                "self.find_non_manifold_vertex('cyl')": set(),
            }
        )

    def test_splitNonManifoldVertex(self):
        """ """
        self.perform_test(
            {
                "self.split_non_manifold_vertex('cyl')": None,
            }
        )

    def test_getNGons(self):
        """ """
        self.perform_test(
            {
                "self.get_ngons('cyl')": [],
            }
        )

    def test_getOverlappingVertices(self):
        """ """
        self.perform_test(
            {
                "self.get_overlapping_vertices('cyl')": [],
            }
        )

    def test_getOverlappingFaces(self):
        """ """
        self.perform_test(
            {
                "self.get_overlapping_faces('cyl')": [],
                "self.get_overlapping_faces('cyl.f[:]')": [],
            }
        )

    def test_getSimilarMesh(self):
        """ """
        self.perform_test(
            {
                "self.get_similar_mesh('cyl')": [],
            }
        )

    def test_getSimilarTopo(self):
        """ """
        self.perform_test(
            {
                "self.get_similar_topo('cyl')": [],
            }
        )


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main(exit=False)


# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

# """
# def test_(self):
#   '''
#   '''
#   self.perform_test({
#       # "self.": '',
#   })


# def test_(self):
#   '''
#   '''
#   self.perform_test({
#       # "self.": '',
#   })


# def test_(self):
#   '''
#   '''
#   self.perform_test({
#       # "self.": '',
#   })


# def test_(self):
#   '''
#   '''
#   self.perform_test({
#       # "self.": '',
#   })
# """

# # Deprecated ---------------------



# !/usr/bin/python
# coding=utf-8
import os
import unittest
import inspect
import pymel.core as pm
from mayatk import NodeUtils


# sfr = pm.melGlobals['cmdScrollFieldReporter']
# pm.cmdScrollFieldReporter(sfr, edit=1, clear=1)


class Main(unittest.TestCase):
    """Main test class."""

    def perform_test(self, cases):
        """Execute the test cases."""
        for case in cases:
            if isinstance(case, str):
                expression = case
                expected_result = cases[case]
                method_name = str(expression).split("(")[0]
            else:
                result, expected_result = case
                method_name = result.__class__.__name__
                expression = None

            try:
                path = os.path.abspath(inspect.getfile(eval(method_name)))
            except (TypeError, IOError):
                path = ""

            if expression:
                result = eval(expression)

            self.assertEqual(
                result,
                expected_result,
                f"\n\n# Error: {path}\n#\tCall: {method_name}({', '.join(map(str, function_args)) if 'function_args' in locals() else ''})\n#\tExpected {type(expected_result)}: {expected_result}\n#\tReturned {type(result)}: {result}",
            )

    @staticmethod
    def replace_mem_address(obj):
        """Replace memory addresses in a string representation of an object with a fixed format of '0x00000000000'.

        Parameters:
                obj (object): The input object. The function first converts this object to a string using the `str` function.

        Returns:
                (str) The string representation of the object with all memory addresses replaced.

        Example:
                >>> replace_mem_address("<class 'str'> <PySide2.QtWidgets.QWidget(0x1ebe2677e80, name='MayaWindow') at 0x000001EBE6D48500>")
                "<class 'str'> <PySide2.QtWidgets.QWidget(0x00000000000, name='MayaWindow') at 0x00000000000>"
        """
        import re

        return re.sub(r"0x[a-fA-F\d]+", "0x00000000000", str(obj))


class NodeUtils_test(Main, NodeUtils):
    """
    set object mode:
            pm.selectMode(object=1)

    set component mode:
            pm.selectMode(component=1)

    set component mode type:
            pm.selectType(allObjects=1)
            pm.selectType(mc=1)
            pm.selectType(vertex=1)
            pm.selectType(edge=1)
            pm.selectType(facet=1)
            pm.selectType(polymeshUV=1)
            pm.selectType(meshUVShell=1)
    """

    # Tear down the any previous test by creating a new scene:
    pm.mel.file(new=True, force=True)

    # assemble the test scene:
    if not pm.objExists("cyl"):
        cyl = pm.polyCylinder(
            radius=5,
            height=10,
            subdivisionsX=12,
            subdivisionsY=1,
            subdivisionsZ=1,
            name="cyl",
        )

    def test_getType(self):
        """ """
        self.assertEqual(self.get_type("cyl"), "transform")
        self.assertEqual(self.get_type("cylShape"), "mesh")
        self.assertEqual(self.get_type("cylShape.vtx[0]"), "vtx")
        self.assertEqual(self.get_type("cylShape.e[0]"), "e")
        self.assertEqual(self.get_type("cylShape.f[0]"), "f")

    def test_getTransformNode(self):
        """ """
        self.assertEqual(self.get_transform_node("cyl"), "cyl")
        self.assertEqual(self.get_transform_node("cylShape"), "cyl")

    def test_getShapeNode(self):
        """ """
        self.assertEqual(self.get_shape_node("cyl"), "cylShape")
        self.assertEqual(self.get_shape_node("cylShape"), "cylShape")

    def test_getHistoryNode(self):
        """ """
        self.assertEqual(self.get_history_node("cyl"), "polyCylinder1")
        self.assertEqual(self.get_history_node("cylShape"), "polyCylinder1")

    def test_isLocator(self):
        """ """
        if not pm.objExists("loc"):
            loc = pm.spaceLocator(name="loc")

        self.assertEqual(self.is_locator("cyl"), False)
        self.assertEqual(self.is_locator(loc), True)

    def test_isGroup(self):
        # Single node tests
        pm.polyCube(name="cube")
        pm.polySphere(name="sphere")
        self.assertEqual(self.is_group("cube"), False)  # Single cube
        self.assertEqual(self.is_group("cubeShape"), False)  # Shape node
        self.assertEqual(self.is_group("sphereShape.vtx[0]"), False)  # Single vertex

        # Group tests
        pm.group("cube", "sphere", name="group1")
        self.assertEqual(self.is_group("group1"), True)  # Group with multiple children

        # Nested group tests
        pm.polyCone(name="cone")
        pm.group("group1", "cone", name="group2")
        self.assertEqual(
            self.is_group("group2"), True
        )  # Group containing another group

        # Empty group tests
        pm.group(name="emptyGroup")
        self.assertEqual(self.is_group("emptyGroup"), True)  # Empty group

        # Cleanup
        pm.delete("group2", "emptyGroup")

    def test_getGroups(self):
        """ """
        self.assertEqual(self.get_groups(), [])

    def test_getParent(self):
        """ """
        self.assertEqual(self.get_parent("cyl"), None)

    def test_getChildren(self):
        """ """
        self.assertEqual(self.get_children("cyl"), [])

    def test_getNodeAttributes(self):
        """ """
        # self.assertEqual(self.get_node_attributes(), None)

    def test_setNodeAttributes(self):
        """ """
        # self.assertEqual(self.set_node_attributes(), None)

    def test_connectAttributes(self):
        """ """
        # self.assertEqual(self.connect_attributes(), None)

    def test_createRenderNode(self):
        """ """
        # self.assertEqual(self.create_render_node(), None)

    def test_getIncomingNodeByType(self):
        """ """
        # self.get_incoming_node_by_type(), None)

    def test_getOutgoingNodeByType(self):
        """ """
        # self.assertEqual(self.get_outgoing_node_by_type(), None)

    def test_connectMultiAttr(self):
        """ """
        # self.assertEqual(self.connect_multi_attr(), None)

    def test_nodeExists(self):
        """ """
        # self.assertEqual(self.node_exists(), None)

    def test_createAssembly(self):
        """ """
        # self.assertEqual(self.create_assembly(), None)


# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main(exit=False)

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

# """

# def test_(self):
#   '''
#   '''
#   self.perform_test({
#       "self.()": None,
#   })

# def test_(self):
#   '''
#   '''
#   self.perform_test({
#       # "self.": '',
#   })

# def test_(self):
#   '''
#   '''
#   self.perform_test({
#       # "self.": '',
#   })

# def test_(self):
#   '''
#   '''
#   self.perform_test({
#       # "self.": '',
#   })
# """

# Deprecated ---------------------



# !/usr/bin/python
# coding=utf-8
try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)
import pythontk as ptk

# from this package:
from mayatk import core_utils, node_utils, component_utils, xform_utils


class EditUtils:
    """ """

    @staticmethod
    @core_utils.CoreUtils.undo
    def rename(objects, to, fltr="", regex=False, ignore_case=False):
        """Rename scene objects.

        Parameters:
            objects (str/obj/list): The object(s to rename. If nothing is given, all scene objects will be renamed.
            to (str): Desired name: An optional asterisk modifier can be used for formatting
                    chars - replace all.
                    *chars* - replace only.
                    *chars - replace suffix.
                    **chars - append suffix.
                    chars* - replace prefix.
                    chars** - append prefix.
            fltr (str): Optionally, filter which the given objects to rename using the following:
                    An asterisk denotes startswith*, *endswith, *contains*, and multiple search strings can be separated by pipe ('|') chars.
                    chars - Search exact.
                    *chars* - Search contains chars.
                    *chars - Search endswith chars.
                    chars* - Search startswith chars.
                    chars|chars - Search any of.  can be used in conjuction with other modifiers.
            regex (bool): If True, regular expression syntax is used instead of the default '*' and '|' modifiers.
            ignore_case (bool): Ignore case when searching. Applies only to the 'fltr' parameter's search.

        Example:
            rename(r'Cube', '*001', regex=True) #replace chars after 'fltr' on any object with a name that contains 'Cube'. ie. 'polyCube001' from 'polyCube'
            rename(r'Cube', '**001', regex=True) #append chars on any object with a name that contains 'Cube'. ie. 'polyCube1001' from 'polyCube1'
        """
        # pm.undoInfo (openChunk=1)
        objects = pm.ls(objectsOnly=1) if not objects else pm.ls(objects)

        # get the short names from the long in order to correctly format. ex. 'NUT_' from: 'CENTER_HINGE_FEMALE_GRP|NUT_'
        long_names = [obj.name() for obj in objects]
        short_names = [
            ii if ii else i for i, ii in ptk.split_at_chars(long_names)
        ]  # split the long names at the last '|' to get the short name.

        names = ptk.find_str_and_format(
            short_names,
            to,
            fltr,
            regex=regex,
            ignore_case=ignore_case,
            return_orig_strings=True,
        )
        print("# Rename: Found {} matches. #".format(len(names)))

        for i, (oldName, newName) in enumerate(names):
            oldName = long_names[
                i
            ]  # use the long name to reference the object instead.
            try:
                if pm.objExists(oldName):
                    n = pm.rename(
                        oldName, newName
                    )  # Rename the object with the new name
                    if not n == newName:
                        print(
                            '# Warning: Attempt to rename "{}" to "{}" failed. Renamed instead to "{}". #'.format(
                                oldName, newName, n
                            )
                        )
                    else:
                        print(
                            '# Result: Successfully renamed "{}" to "{}". #'.format(
                                oldName, newName
                            )
                        )

            except Exception as e:
                if not pm.ls(oldName, readOnly=True) == []:  # ignore read-only errors.
                    print(
                        '# Error: Attempt to rename "{}" to "{}" failed. {} #'.format(
                            oldName, newName, str(e).rstrip()
                        )
                    )
        # pm.undoInfo (closeChunk=1)

    @staticmethod
    @core_utils.CoreUtils.undo
    def set_case(objects=[], case="caplitalize"):
        """Rename objects following the given case.

        Parameters:
            objects (str/list): The objects to rename. default:all scene objects
            case (str): Desired case using python case operators.
                    valid: 'upper', 'lower', 'caplitalize', 'swapcase' 'title'. default:'caplitalize'
        Example:
            set_case(pm.ls(sl=1), 'upper')
        """
        # pm.undoInfo(openChunk=1)
        for obj in pm.ls(objects):
            name = obj.name()

            newName = getattr(name, case)()
            try:
                pm.rename(name, newName)
            except Exception as error:
                if not pm.ls(obj, readOnly=True) == []:  # Ignore read-only errors.
                    print(name + ": ", error)
        # pm.undoInfo(closeChunk=1)

    @staticmethod
    @core_utils.CoreUtils.undo
    def append_location_based_suffix(
        objects,
        alphanumeric=False,
        strip_trailing_ints=True,
        strip_trailing_alpha=True,
        reverse=False,
    ):
        """Rename objects with a suffix defined by its location from origin.

        Parameters:
            objects (str)(int/list): The object(s) to rename.
            alphanumeric (str): When True use an alphanumeric character as a suffix when there is less than 26 objects else use integers.
            strip_trailing_ints (bool): Strip any trailing integers. ie. 'cube123'
            strip_trailing_alpha (bool): Strip any trailing uppercase alphanumeric chars that are prefixed with an underscore.  ie. 'cube_A'
            reverse (bool): Reverse the naming order. (Farthest object first)
        """
        import string
        import re

        length = len(objects)
        if alphanumeric:
            if length <= 26:
                suffix = string.ascii_lowercase.upper()
        else:
            suffix = [str(n).zfill(len(str(length))) for n in range(length)]

        ordered_objs = xform_utils.XformUtils.order_by_distance(
            objects, reverse=reverse
        )

        newNames = {}  # the object with the new name set as a key.
        for n, obj in enumerate(ordered_objs):
            current_name = obj.name()

            while (
                (current_name[-1] == "_" or current_name[-1].isdigit())
                and strip_trailing_ints
            ) or (
                (
                    len(current_name) > 1
                    and current_name[-2] == "_"
                    and current_name[-1].isupper()
                )
                and strip_trailing_alpha
            ):
                if (
                    current_name[-2] == "_" and current_name[-1].isupper()
                ) and strip_trailing_alpha:  # trailing underscore and uppercase alphanumeric char.
                    current_name = re.sub(
                        re.escape(current_name[-2:]) + "$", "", current_name
                    )

                if (
                    current_name[-1] == "_" or current_name[-1].isdigit()
                ) and strip_trailing_ints:  # trailing underscore and integers.
                    current_name = re.sub(
                        re.escape(current_name[-1:]) + "$", "", current_name
                    )

            newNames[obj] = current_name + "_" + suffix[n]

        for (
            obj
        ) in (
            ordered_objs
        ):  # rename all with a placeholder first so that there are no conflicts.
            pm.rename(obj, "p0000000000")
        for obj in ordered_objs:  # rename all with the new names.
            pm.rename(obj, newNames[obj])

    @staticmethod
    def snap_closest_verts(obj1, obj2, tolerance=10.0, freeze_transforms=False):
        """Snap the vertices from object one to the closest verts on object two.

        Parameters:
            obj1 (obj): The object in which the vertices are moved from.
            obj2 (obj): The object in which the vertices are moved to.
            tolerance (float) = Maximum search distance.
            freeze_transforms (bool): Reset the selected transform and all of its children down to the shape level.
        """
        vertices = component_utils.ComponentUtils.get_components(obj1, "vertices")
        closestVerts = component_utils.ComponentUtils.get_closest_vertex(
            vertices, obj2, tolerance=tolerance, freeze_transforms=freeze_transforms
        )

        progressBar = "mainProgressBar"
        pm.progressBar(
            progressBar,
            edit=True,
            beginProgress=True,
            isInterruptable=True,
            status="Snapping Vertices ...",
            maxValue=len(closestVerts),
        )

        pm.undoInfo(openChunk=True)
        for v1, v2 in closestVerts.items():
            if pm.progressBar(progressBar, query=True, isCancelled=True):
                break

            v2Pos = pm.pointPosition(v2, world=True)
            pm.xform(v1, translation=v2Pos, worldSpace=True)

            pm.progressBar(progressBar, edit=True, step=1)
        pm.undoInfo(closeChunk=True)

        pm.progressBar(progressBar, edit=True, endProgress=True)

    @staticmethod
    def merge_vertices(objects, selected=False, tolerance=0.001):
        """Merge Vertices on the given objects.

        Parameters:
            objects (str/obj/list): The object(s) to merge vertices on.
            selected (bool): Merge only the currently selected components.
            tolerance (float) = The maximum merge distance.
        """
        for obj in pm.ls(objects):
            if selected:  # merge selected components.
                if pm.filterExpand(selectionMask=31):  # selectionMask=vertices
                    sel = pm.ls(obj, sl=1)
                    pm.polyMergeVertex(
                        sel,
                        distance=tolerance,
                        alwaysMergeTwoVertices=True,
                        constructionHistory=True,
                    )
                else:  # if selection type =edges or facets:
                    pm.mel.MergeToCenter()

            else:  # merge all vertices on the object.
                vertices = obj.vtx[:]  # mel expression: select -r geometry.vtx[0:1135];
                pm.polyMergeVertex(
                    vertices,
                    distance=tolerance,
                    alwaysMergeTwoVertices=False,
                    constructionHistory=False,
                )
                # return to original state
                pm.select(clear=1)
                pm.select(objects)

    @staticmethod
    def get_all_faces_on_axis(obj, axis="-x", localspace=False):
        """Get all faces on a specified axis.

        Parameters:
            obj (str/obj): The name of the geometry.
            axis (str): The representing axis. case insensitive. (valid: 'x', '-x', 'y', '-y', 'z', '-z')
            localspace (bool): Specify world or local space.

        Example:
            get_all_faces_on_axis('polyObject', 'y')
        """
        obj = pm.ls(obj)[0] if pm.ls(obj) else None

        if not obj:
            raise ValueError(f"No such object exists: {obj}")

        # Get shape nodes
        shapes = obj.getShapes()
        if not shapes:
            raise ValueError("The object has no shape nodes.")

        # Validate axis
        valid_axes = {"x": 0, "-x": 0, "y": 1, "-y": 1, "z": 2, "-z": 2}
        axis = axis.lower()
        if axis not in valid_axes:
            raise ValueError(
                "Invalid axis. Valid options are 'x', '-x', 'y', '-y', 'z', '-z'"
            )

        i = valid_axes[axis]
        # Collect faces from all geometry type shapes
        faces = []
        for shape in shapes:
            if pm.nodeType(shape) not in ["mesh", "nurbsSurface", "subdiv"]:
                continue
            faces.extend(pm.filterExpand(f"{shape.name()}.f[*]", sm=34))

        if not faces:
            raise ValueError("The shape nodes have no faces.")

        # Get faces on the specified axis
        if axis.startswith("-"):
            return [
                face for face in faces if pm.exactWorldBoundingBox(face)[i] < -0.00001
            ]
        else:
            return [
                face for face in faces if pm.exactWorldBoundingBox(face)[i] > -0.00001
            ]

    @classmethod
    def delete_along_axis(cls, objects, axis="-x"):
        """Delete components of the given mesh object along the specified axis.

        Parameters:
            obj (obj): Mesh object.
            axis (str): Axis to delete on. ie. '-x' Components belonging to the mesh object given in the 'obj' arg, that fall on this axis, will be deleted.
        """
        # Get any mesh type child nodes of obj.
        for node in (o for o in pm.ls(objects) if not node_utils.NodeUtils.is_group(o)):
            faces = cls.get_all_faces_on_axis(node, axis)
            # If all faces fall on the specified axis.
            if len(faces) == pm.polyEvaluate(node, face=1):
                pm.delete(node)  # Delete entire node
            else:
                pm.delete(faces)  # Else, delete any individual faces.

    @classmethod
    def clean_geometry(
        cls,
        objects,
        allMeshes=False,
        repair=False,
        quads=False,
        nsided=False,
        concave=False,
        holed=False,
        nonplanar=False,
        zeroGeom=False,
        zeroGeomTol=0.000010,
        zeroEdge=False,
        zeroEdgeTol=0.000010,
        zeroMap=False,
        zeroMapTol=0.000010,
        sharedUVs=False,
        nonmanifold=False,
        lamina=False,
        invalidComponents=False,
        split_non_manifold_vertex=False,
        historyOn=True,
    ):
        """Select or remove unwanted geometry from a polygon mesh.

        Parameters:
            objects (str/obj/list): The polygon objects to clean.
            allMeshes (bool): Clean all geomtry in the scene instead of only the current selection.
            repair (bool): Attempt to repair instead of just selecting geometry.
        """
        arg_list = '"{0}","{1}","{2}","{3}","{4}","{5}","{6}","{7}","{8}","{9}","{10}","{11}","{12}","{13}","{14}","{15}","{16}","{17}"'.format(
            allMeshes,
            1 if repair else 2,
            historyOn,
            quads,
            nsided,
            concave,
            holed,
            nonplanar,
            zeroGeom,
            zeroGeomTol,
            zeroEdge,
            zeroEdgeTol,
            zeroMap,
            zeroMapTol,
            sharedUVs,
            nonmanifold,
            lamina,
            invalidComponents,
        )
        command = (
            "polyCleanupArgList 4 {" + arg_list + "}"
        )  # command = 'polyCleanup '+arg_list #(not used because of arg count error, also the quotes in the arg list would need to be removed).

        if split_non_manifold_vertex:  # Split Non-Manifold Vertex
            nonManifoldVerts = cls.find_non_manifold_vertex(
                objects, select=2
            )  # Select: 0=off, 1=on, 2=on while keeping any existing vertex selections. (default: 1)
            if repair:
                for vertex in nonManifoldVerts:
                    cls.split_non_manifold_vertex(
                        vertex, select=True
                    )  # select(bool): Select the vertex after the operation. (default: True)

        pm.select(objects)
        pm.mel.eval(command)
        # print (command)

    @staticmethod
    def get_overlapping_duplicates(
        objects=[], retain_given_objects=False, select=False, verbose=False
    ):
        """Find any duplicate overlapping geometry at the object level.

        Parameters:
            objects (list): A list of objects to find duplicate overlapping geometry for. Default is selected objects, or all if nothing is selected.
            retain_given_objects (bool): Search only for duplicates of the given objects (or any selected objects if None given), and omit them from the return results.
            select (bool): Select any found duplicate objects.
            verbose (bool): Print each found object to console.

        Returns:
            (set)

        Example:
            duplicates = get_overlapping_duplicates(retain_given_objects=True, select=True, verbose=True)
        """
        scene_objs = pm.ls(transforms=1, geometry=1)  # get all scene geometry

        # attach a unique identifier consisting each objects polyEvaluate attributes, and it's bounding box center point in world space.
        scene_objs = {
            i: str(pm.objectCenter(i)) + str(pm.polyEvaluate(i))
            for i in scene_objs
            if not node_utils.NodeUtils.is_group(i)
        }
        selected_objs = pm.ls(scene_objs.keys(), sl=1) if not objects else objects

        objs_inverted = {}  # invert the dict, combining objects with like identifiers.
        for k, v in scene_objs.items():
            objs_inverted[v] = objs_inverted.get(v, []) + [k]

        duplicates = set()
        for k, v in objs_inverted.items():
            if len(v) > 1:
                if selected_objs:  # limit scope to only selected objects.
                    if set(selected_objs) & set(
                        v
                    ):  # if any selected objects in found duplicates:
                        if retain_given_objects:
                            [
                                duplicates.add(i) for i in v if i not in selected_objs
                            ]  # add any duplicated of that object, omitting the selected object.
                        else:
                            [
                                duplicates.add(i) for i in v[1:]
                            ]  # add all but the first object to the set of duplicates.
                else:
                    [
                        duplicates.add(i) for i in v[1:]
                    ]  # add all but the first object to the set of duplicates.

        if verbose:
            for i in duplicates:
                print("# Found: overlapping duplicate object: {} #".format(i))
        print("# {} overlapping duplicate objects found. #".format(len(duplicates)))

        if select:
            pm.select(duplicates)

        return duplicates

    @staticmethod
    def find_non_manifold_vertex(objects, select=1):
        """Locate a connected vertex of non-manifold geometry where the faces share a single vertex.

        Parameters:
            objects (str/obj/list): A polygon mesh, or a list of meshes.
            select (int): Select any found non-manifold vertices. 0=off, 1=on, 2=on while keeping any existing vertex selections. (default: 1)

        Returns:
            (set) any found non-manifold verts.
        """
        pm.undoInfo(openChunk=True)
        nonManifoldVerts = set()

        vertices = component_utils.ComponentUtils.get_components(objects, "vertices")
        for vertex in vertices:
            connected_faces = pm.polyListComponentConversion(
                vertex, fromVertex=1, toFace=1
            )  # pm.mel.PolySelectConvert(1) #convert to faces
            connected_faces_flat = pm.ls(
                connected_faces, flatten=1
            )  # selectedFaces = pm.ls(sl=1, flatten=1)

            # get a list of the edges of each face that is connected to the original vertex.
            edges_sorted_by_face = []
            for face in connected_faces_flat:
                connected_edges = pm.polyListComponentConversion(
                    face, fromFace=1, toEdge=1
                )  # pm.mel.PolySelectConvert(1) #convert to faces
                connected_edges_flat = [
                    str(i) for i in pm.ls(connected_edges, flatten=1)
                ]  # selectedFaces = pm.ls(sl=1, flatten=1)
                edges_sorted_by_face.append(connected_edges_flat)

            out = (
                []
            )  # 1) take first set A from list. 2) for each other set B in the list do if B has common element(s) with A join B into A; remove B from list. 3) repeat 2. until no more overlap with A. 4) put A into outpup. 5) repeat 1. with rest of list.
            while len(edges_sorted_by_face) > 0:
                first, rest = (
                    edges_sorted_by_face[0],
                    edges_sorted_by_face[1:],
                )  # first list, all other lists, of the list of lists.
                first = set(first)

                lf = -1
                while len(first) > lf:
                    lf = len(first)

                    rest2 = []
                    for r in rest:
                        if len(first.intersection(set(r))) > 0:
                            first |= set(r)
                        else:
                            rest2.append(r)
                    rest = rest2

                out.append(first)
                edges_sorted_by_face = rest

            if len(out) > 1:
                nonManifoldVerts.add(vertex)
        pm.undoInfo(closeChunk=True)

        if select == 2:
            pm.select(nonManifoldVerts, add=1)
        elif select == 1:
            pm.select(nonManifoldVerts)

        return nonManifoldVerts

    @staticmethod
    def split_non_manifold_vertex(vertex, select=True):
        """Separate a connected vertex of non-manifold geometry where the faces share a single vertex.

        Parameters:
            vertex (str/obj): A single polygon vertex.
            select (bool): Select the vertex after the operation. (default is True)
        """
        pm.undoInfo(openChunk=True)
        connected_faces = pm.polyListComponentConversion(
            vertex, fromVertex=1, toFace=1
        )  # pm.mel.PolySelectConvert(1) #convert to faces
        connected_faces_flat = pm.ls(
            connected_faces, flatten=1
        )  # selectedFaces = pm.ls(sl=1, flatten=1)

        pm.polySplitVertex(vertex)

        # get a list for the vertices of each face that is connected to the original vertex.
        verts_sorted_by_face = []
        for face in connected_faces_flat:
            connected_verts = pm.polyListComponentConversion(
                face, fromFace=1, toVertex=1
            )  # pm.mel.PolySelectConvert(1) #convert to faces
            connected_verts_flat = [
                str(i) for i in pm.ls(connected_verts, flatten=1)
            ]  # selectedFaces = pm.ls(sl=1, flatten=1)
            verts_sorted_by_face.append(connected_verts_flat)

        # 1) take first set A from list. 2) for each other set B in the list do if B has common element(s) with A join B into A; remove B from list. 3) repeat 2. until no more overlap with A. 4) put A into outpup. 5) repeat 1. with rest of list.
        out = []
        while len(verts_sorted_by_face) > 0:
            # first, *rest = verts_sorted_by_face
            first, rest = (
                verts_sorted_by_face[0],
                verts_sorted_by_face[1:],
            )
            first = set(first)

            lf = -1
            while len(first) > lf:
                lf = len(first)

                rest2 = []
                for r in rest:
                    if len(first.intersection(set(r))) > 0:
                        first |= set(r)
                    else:
                        rest2.append(r)
                rest = rest2

            out.append(first)
            verts_sorted_by_face = rest

        for vertex_set in out:
            pm.polyMergeVertex(vertex_set, distance=0.001)

        # deselect the vertices that were selected during the polyMergeVertex operation.
        pm.select(vertex_set, deselect=1)
        if select:
            pm.select(vertex, add=1)
        pm.undoInfo(closeChunk=True)

    @staticmethod
    def get_ngons(objects, repair=False):
        """Get any N-Gons from the given object using selection contraints.

        Parameters:
            objects (str/obj/list): The objects to query.
            repair (bool): Repair any found N-gons.

        Returns:
            (list)
        """
        pm.select(objects)
        # Change to Component mode to retain object highlighting
        pm.mel.changeSelectMode(1)
        # Change to Face Component Mode
        pm.selectType(smp=0, sme=1, smf=0, smu=0, pv=0, pe=1, pf=0, puv=0)
        # Select Object/s and Run Script to highlight N-Gons
        pm.polySelectConstraint(mode=3, type=0x0008, size=3)
        nGons = pm.ls(sl=1)
        pm.polySelectConstraint(disable=1)

        if repair:  # convert N-Sided Faces To Quads
            pm.polyQuad(nGons, angle=30, kgb=1, ktb=1, khe=1, ws=1)

        return nGons

    @staticmethod
    def get_overlapping_vertices(objects, threshold=0.0003):
        """Query the given objects for overlapping vertices.

        Parameters:
            objects (str/obj/list): The objects to query.
            threshold (float) = The maximum allowed distance.

        Returns:
            (list)
        """
        import maya.OpenMaya as om

        result = []
        for mfnMesh in core_utils.CoreUtils.mfn_mesh_generator(objects):
            points = om.MPointArray()
            mfnMesh.getPoints(points, om.MSpace.kWorld)

            for i in range(points.length()):
                for ii in range(points.length()):
                    if i == ii:
                        continue

                    dist = points[i].distanceTo(points[ii])
                    if dist < threshold:
                        if i not in result:
                            result.append(i)

                        if ii not in result:
                            result.append(ii)
        return result

    @classmethod
    def get_overlapping_faces(cls, objects):
        """Get any duplicate overlapping faces of the given objects.

        :Parameters:
            objects (str/obj/list): Faces or polygon objects.

        Returns:
            (list) duplicate overlapping faces.

        Example: pm.select(get_overlapping_faces(selection))
        """
        if not objects:
            return []

        elif not pm.nodeType(objects) == "mesh":  # if the objects are not faces.
            duplicates = ptk.flatten(
                [
                    cls.get_overlapping_faces(obj.faces)
                    for obj in pm.ls(objects, objectsOnly=1)
                ]
            )
            return list(duplicates)

        face, *otherFaces = pm.ls(objects)
        face_vtx_positions = [
            v.getPosition()
            for v in pm.ls(pm.polyListComponentConversion(face, toVertex=1), flatten=1)
        ]

        duplicates = []
        for otherFace in otherFaces:
            otherFace_vtx_positions = [
                v.getPosition()
                for v in pm.ls(
                    pm.polyListComponentConversion(otherFace, toVertex=1), flatten=1
                )
            ]

            if face_vtx_positions == otherFace_vtx_positions:  # duplicate found.
                duplicates.append(otherFace)
                otherFaces.remove(otherFace)

        if otherFaces:
            # after adding any found duplicates, call again with any remaining faces.
            duplicates += cls.get_overlapping_faces(otherFaces)

        return duplicates

    @staticmethod
    def get_similar_mesh(obj, tolerance=0.0, inc_orig=False, **kwargs):
        """Find similar geometry objects using the polyEvaluate command.
        Default behaviour is to compare all flags.

        Parameters:
            obj (str/obj/list): The object to find similar for.
            tolerance (float) = The allowed difference in any of the given polyEvalute flag results (that return an int, float (or list of the int or float) value(s)).
            inc_orig (bool): Include the original given obj with the return results.
            kwargs (bool): Any keyword argument 'polyEvaluate' takes. Used to filter the results.
                    ex: vertex, edge, face, uvcoord, triangle, shell, boundingBox, boundingBox2d,
                    vertexComponent, boundingBoxComponent, boundingBoxComponent2d, area, worldArea
        Returns:
            (list) Similar objects.

        Example:
            get_similar_mesh(selection, vertex=1, area=1)
        """
        lst = (
            lambda x: list(x)
            if isinstance(x, (list, tuple, set))
            else list(x.values())
            if isinstance(x, dict)
            else [x]
        )  # assure the returned result from polyEvaluate is a list of values.

        obj, *other = pm.ls(obj, long=True, transforms=True)
        objProps = lst(pm.polyEvaluate(obj, **kwargs))

        otherSceneMeshes = set(
            pm.filterExpand(pm.ls(long=True, typ="transform"), selectionMask=12)
        )  # polygon selection mask.
        similar = pm.ls(
            [
                m
                for m in otherSceneMeshes
                if ptk.are_similar(
                    objProps, lst(pm.polyEvaluate(m, **kwargs)), tolerance=tolerance
                )
                and m != obj
            ]
        )
        return similar + [obj] if inc_orig else similar

    @staticmethod
    def get_similar_topo(obj, inc_orig=False, **kwargs):
        """Find similar geometry objects using the polyCompare command.
        Default behaviour is to compare all flags.

        Parameters:
            obj (str/obj/list): The object to find similar for.
            inc_orig (bool): Include the original given obj with the return results.
            kwargs (bool): Any keyword argument 'polyCompare' takes. Used to filter the results.
                    ex: vertices, edges, faceDesc, uvSets, uvSetIndices, colorSets, colorSetIndices, userNormals
        Returns:
            (list) Similar objects.
        """
        obj, *other = pm.filterExpand(
            pm.ls(obj, long=True, tr=True), selectionMask=12
        )  # polygon selection mask.

        otherSceneMeshes = set(
            pm.filterExpand(pm.ls(long=True, typ="transform"), sm=12)
        )
        similar = pm.ls(
            [
                m
                for m in otherSceneMeshes
                if pm.polyCompare(obj, m, **kwargs) == 0 and m != obj
            ]
        )  # 0:equal,Verts:1,Edges:2,Faces:4,UVSets:8,UVIndices:16,ColorSets:32,ColorIndices:64,UserNormals=128. So a return value of 3 indicates both vertices and edges are different.
        return similar + [obj] if inc_orig else similar


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    pass

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

# deprecated ---------------------
